import chess
import torch

import chipiron.environments.chess_env.board as boards
from chipiron.environments.chess_env.board.iboard import IBoard

from .board_representation import Representation364


def create_from_board_and_from_parent(
    board: IBoard,
    board_modifications: boards.BoardModificationP,
    parent_node_board_representation: Representation364,
) -> Representation364:
    """
    Converts the node, board modifications, and parent node into a tensor representation.

    Args:
        board (IBoard): The current board in the tree.
        board_modifications (board_mod.BoardModification): The modifications made to the board.
        parent_node (AlgorithmNode): The parent node of the current node.

    Returns:
        Representation364: The tensor representation of the node, board modifications, and parent node.
    """

    assert isinstance(parent_node_board_representation, Representation364)
    tensor_white = torch.empty_like(
        parent_node_board_representation.tensor_white
    ).copy_(parent_node_board_representation.tensor_white)
    tensor_black = torch.empty_like(
        parent_node_board_representation.tensor_black
    ).copy_(parent_node_board_representation.tensor_black)

    for removal in board_modifications.removals:
        piece_type = removal.piece
        piece_color = removal.color
        square = removal.square
        piece_code = piece_type - 1
        if piece_color == chess.BLACK:
            square_index = chess.square_mirror(square)
            index = 64 * piece_code + square_index
            tensor_black[index] = 0
        else:
            square_index = square
            index = 64 * piece_code + square_index
            tensor_white[index] = 0

    for appearance in board_modifications.appearances:
        # print('app',appearance)
        piece_type = appearance.piece
        piece_color = appearance.color
        square = appearance.square
        piece_code = piece_type - 1
        if piece_color == chess.BLACK:
            square_index = chess.square_mirror(square)
            index = 64 * piece_code + square_index
            tensor_black[index] = 1
        else:
            square_index = square
            index = 64 * piece_code + square_index
            tensor_white[index] = 1

    tensor_castling_white = torch.zeros(2, requires_grad=False, dtype=torch.float)
    tensor_castling_black = torch.zeros(2, requires_grad=False, dtype=torch.float)

    tensor_castling_white[0] = bool(board.castling_rights & chess.BB_A1)
    tensor_castling_white[1] = bool(board.castling_rights & chess.BB_H1)
    tensor_castling_black[0] = bool(board.castling_rights & chess.BB_A8)
    tensor_castling_black[1] = bool(board.castling_rights & chess.BB_H8)

    representation = Representation364(
        tensor_white=tensor_white,
        tensor_black=tensor_black,
        tensor_castling_black=tensor_castling_black,
        tensor_castling_white=tensor_castling_white,
    )

    return representation


def create_from_board(board: boards.IBoard) -> Representation364:
    """
    Create a Representation364 object from a board.

    Args:
        board (BoardChi): The chess board.

    Returns:
        Representation364: The created Representation364 object.
    """

    white: torch.Tensor
    black: torch.Tensor
    castling_white: torch.Tensor
    castling_black: torch.Tensor
    white, black, castling_black, castling_white = d()

    p: dict[chess.Square, tuple[int, bool]] = e(board)
    a(black, white, p)
    b(board, castling_black, castling_white)
    representation = c(castling_black, castling_white, black, white)

    return representation


def e(board: boards.IBoard) -> dict[chess.Square, tuple[int, bool]]:
    piece_map: dict[chess.Square, tuple[int, bool]] = board.piece_map()
    return piece_map


def d() -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    white: torch.Tensor = torch.zeros(384, dtype=torch.float)
    black: torch.Tensor = torch.zeros(384, dtype=torch.float)
    castling_white: torch.Tensor = torch.zeros(2, dtype=torch.float)
    castling_black: torch.Tensor = torch.zeros(2, dtype=torch.float)
    return white, black, castling_black, castling_white


def c(
    castling_black: torch.Tensor,
    castling_white: torch.Tensor,
    black: torch.Tensor,
    white: torch.Tensor,
) -> Representation364:
    representation: Representation364 = Representation364(
        tensor_white=white,
        tensor_black=black,
        tensor_castling_black=castling_black,
        tensor_castling_white=castling_white,
    )
    return representation


def b(
    board: boards.IBoard, castling_black: torch.Tensor, castling_white: torch.Tensor
) -> None:
    castling_white[0] = board.has_queenside_castling_rights(chess.WHITE)
    castling_white[1] = board.has_kingside_castling_rights(chess.WHITE)
    castling_black[0] = board.has_queenside_castling_rights(chess.BLACK)
    castling_black[1] = board.has_kingside_castling_rights(chess.BLACK)


def a(
    black: torch.Tensor,
    white: torch.Tensor,
    piece_map: dict[chess.Square, tuple[int, bool]],
) -> None:
    square: chess.Square
    piece: tuple[int, bool]
    for square, piece in piece_map.items():
        piece_code: int = piece[0] - 1
        piece_color: bool = piece[1]
        square_index: chess.Square
        index: int
        if piece_color == chess.BLACK:
            square_index = chess.square_mirror(square)
            index = 64 * piece_code + square_index
            black[index] = 1.0
        else:
            square_index = square
            index = 64 * piece_code + square_index
            white[index] = 1.0



"""
Module that contains the NeuralNetBoardEvalArgs class.
"""

from dataclasses import dataclass, field
from typing import Literal

from chipiron.players.boardevaluators.board_evaluation.board_evaluation import (
    PointOfView,
)
from chipiron.players.boardevaluators.board_evaluator_type import BoardEvalTypes
from chipiron.players.boardevaluators.neural_networks.factory import (
    NeuralNetModelsAndArchitecture,
)
from chipiron.players.boardevaluators.neural_networks.input_converters.ModelInputRepresentationType import (
    ModelInputRepresentationType,
)
from chipiron.players.boardevaluators.neural_networks.models.multi_layer_perceptron import (
    MultiLayerPerceptronArgs,
)
from chipiron.players.boardevaluators.neural_networks.neural_net_architecture_args import (
    NeuralNetArchitectureArgs,
)
from chipiron.players.boardevaluators.neural_networks.NNModelType import (
    ActivationFunctionType,
    NNModelType,
)
from chipiron.players.boardevaluators.neural_networks.output_converters.model_output_type import (
    ModelOutputType,
)


@dataclass
class NeuralNetBoardEvalArgs:
    """
    NeuralNetBoardEvalArgs encapsulates the configuration arguments required for evaluating board positions
    using a neural network-based node evaluator.

    Attributes:
        neural_nets_model_and_architecture (NeuralNetModelsAndArchitecture):
            Specifies the neural network model, its architecture, and associated parameters.
            Defaults to a MultiLayerPerceptron with predefined layer sizes and activation functions.
        type (Literal[NodeEvaluatorTypes.NeuralNetwork]):
            The type of node evaluator, which must be set to 'NeuralNetwork'.
    """

    type: Literal[BoardEvalTypes.NEURAL_NET_BOARD_EVAL]
    neural_nets_model_and_architecture: NeuralNetModelsAndArchitecture = field(
        default_factory=lambda: NeuralNetModelsAndArchitecture(
            model_weights_file_name="*default*",
            nn_architecture_args=NeuralNetArchitectureArgs(
                model_type_args=MultiLayerPerceptronArgs(
                    type=NNModelType.MULTI_LAYER_PERCEPTRON,
                    number_neurons_per_layer=[5, 1],
                    list_of_activation_functions=[
                        ActivationFunctionType.TANGENT_HYPERBOLIC
                    ],
                ),
                model_output_type=ModelOutputType(
                    point_of_view=PointOfView.PLAYER_TO_MOVE
                ),
                model_input_representation_type=ModelInputRepresentationType.PIECE_DIFFERENCE,
            ),
        )
    )

    def __post_init__(self) -> None:
        """
        Performs additional initialization after the object is created.

        Raises:
            ValueError: If the type is not NodeEvaluatorTypes.NeuralNetwork.
        """
        if self.type != BoardEvalTypes.NEURAL_NET_BOARD_EVAL:
            raise ValueError("Expecting neural_network as name")
        if (
            self.neural_nets_model_and_architecture.model_weights_file_name
            == "*default*"
        ):
            raise ValueError(f"Expecting a path_to_nn_folder in {__name__}")



"""
This module provides functionality for evaluating nodes in a tree structure.

The module includes a factory function for creating node evaluators, as well as classes for representing
node evaluators, evaluation queries, and node evaluator arguments.

Available objects:
- AllNodeEvaluatorArgs: A named tuple representing all the arguments for creating a node evaluator.
- NodeEvaluator: A class representing a node evaluator.
- create_node_evaluator: A factory function for creating a node evaluator.
- EvaluationQueries: An enumeration representing different types of evaluation queries.
- NodeEvaluatorArgs: A class representing the arguments for creating a node evaluator.
"""

from .factory import create_node_evaluator
from .node_evaluator import EvaluationQueries, NodeEvaluator
from .node_evaluator_args import NodeEvaluatorArgs

__all__ = [
    "NodeEvaluator",
    "create_node_evaluator",
    "EvaluationQueries",
    "NodeEvaluatorArgs",
]



"""
This module provides functions for creating a TreeAndValueMoveSelector object.

The TreeAndValueMoveSelector is a player that uses a tree-based approach to select moves in a game. It evaluates the
game tree using a node evaluator and selects moves based on a set of criteria defined by the node selector. The player
uses a stopping criterion to determine when to stop the search and a recommender rule to recommend a move after
exploration.

This module also provides functions for creating the necessary components of the TreeAndValueMoveSelector, such as the
node evaluator, node selector, tree factory, and tree manager.

"""

import queue
import random
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal

import chipiron.players.move_selector.treevalue.search_factory as search_factories
from chipiron.players.boardevaluators.neural_networks.input_converters.representation_factory_factory import (
    create_board_representation_factory,
)
from chipiron.players.boardevaluators.table_base.factory import AnySyzygyTable
from chipiron.players.move_selector.move_selector_types import MoveSelectorTypes
from chipiron.players.move_selector.treevalue import node_factory
from chipiron.players.move_selector.treevalue.node_evaluator.node_evaluator_args import (
    NodeEvaluatorArgs,
)
from chipiron.players.move_selector.treevalue.progress_monitor.progress_monitor import (
    AllStoppingCriterionArgs,
)
from chipiron.utils.dataclass import IsDataclass

from . import node_evaluator as node_eval
from . import node_selector as node_selector_m
from . import recommender_rule
from . import tree_manager as tree_man
from .indices.node_indices.index_types import IndexComputationType
from .tree_and_value_move_selector import TreeAndValueMoveSelector
from .trees.factory import MoveAndValueTreeFactory

if TYPE_CHECKING:
    from chipiron.players.boardevaluators.neural_networks.input_converters.factory import (
        RepresentationFactory,
    )


@dataclass
class TreeAndValuePlayerArgs:
    """
    Data class for the arguments of a TreeAndValueMoveSelector.
    """

    type: Literal[MoveSelectorTypes.TreeAndValue]  # for serialization
    node_selector: node_selector_m.AllNodeSelectorArgs
    opening_type: node_selector_m.OpeningType
    node_evaluator: NodeEvaluatorArgs
    stopping_criterion: AllStoppingCriterionArgs
    recommender_rule: recommender_rule.AllRecommendFunctionsArgs
    index_computation: IndexComputationType | None = None


def create_tree_and_value_builders(
    args: TreeAndValuePlayerArgs,
    syzygy: AnySyzygyTable | None,
    random_generator: random.Random,
    queue_progress_player: queue.Queue[IsDataclass] | None,
) -> TreeAndValueMoveSelector:
    """
    Create a TreeAndValueMoveSelector object with the given arguments.

    Args:
        args (TreeAndValuePlayerArgs): The arguments for creating the TreeAndValueMoveSelector.
        syzygy (SyzygyTable | None): The SyzygyTable object for tablebase endgame evaluation.
        random_generator (random.Random): The random number generator.

    Returns:
        TreeAndValueMoveSelector: The created TreeAndValueMoveSelector object.

    """

    node_evaluator: node_eval.NodeEvaluator = node_eval.create_node_evaluator(
        arg_board_evaluator=args.node_evaluator, syzygy=syzygy
    )

    # node_factory_name: str = args['node_factory_name'] if 'node_factory_name' in args else 'Base'
    node_factory_name: str = "Base_with_algorithm_tree_node"

    tree_node_factory: node_factory.Base[Any] = node_factory.create_node_factory(
        node_factory_name=node_factory_name
    )

    board_representation_factory: RepresentationFactory[Any] | None
    board_representation_factory = create_board_representation_factory(
        internal_tensor_representation_type=args.node_evaluator.internal_representation_type
    )

    search_factory: search_factories.SearchFactoryP = search_factories.SearchFactory(
        node_selector_args=args.node_selector,
        opening_type=args.opening_type,
        random_generator=random_generator,
        index_computation=args.index_computation,
    )

    algorithm_node_factory: node_factory.AlgorithmNodeFactory
    algorithm_node_factory = node_factory.AlgorithmNodeFactory(
        tree_node_factory=tree_node_factory,
        board_representation_factory=board_representation_factory,
        exploration_index_data_create=search_factory.node_index_create,
    )

    tree_factory = MoveAndValueTreeFactory(
        node_factory=algorithm_node_factory, node_evaluator=node_evaluator
    )

    tree_manager: tree_man.AlgorithmNodeTreeManager
    tree_manager = tree_man.create_algorithm_node_tree_manager(
        algorithm_node_factory=algorithm_node_factory,
        node_evaluator=node_evaluator,
        index_computation=args.index_computation,
        index_updater=search_factory.create_node_index_updater(),
    )

    tree_move_selector: TreeAndValueMoveSelector = TreeAndValueMoveSelector(
        tree_manager=tree_manager,
        random_generator=random_generator,
        tree_factory=tree_factory,
        node_selector_create=search_factory.create_node_selector_factory(),
        stopping_criterion_args=args.stopping_criterion,
        recommend_move_after_exploration=args.recommender_rule,
        queue_progress_player=queue_progress_player,
    )
    return tree_move_selector



"""
Module for the SyzygyTable class.
"""

import chess.syzygy

import chipiron.environments.chess_env.board as boards
from chipiron.players.boardevaluators.table_base import SyzygyTable
from chipiron.utils import path


class SyzygyChiTable(SyzygyTable[boards.BoardChi]):
    """
    A class representing a Syzygy tablebase for chess endgame analysis.

    Attributes:
        table_base (chess.syzygy.Tablebase): The Syzygy tablebase object.

    Methods:
        fast_in_table(board: boards.BoardChi) -> bool:
            Check if the given board is suitable for fast tablebase lookup.

        in_table(board: boards.BoardChi) -> bool:
            Check if the given board is in the tablebase.

        get_over_event(board: boards.BoardChi) -> tuple[Winner, HowOver]:
            Get the winner and how the game is over for the given board.

        val(board: boards.BoardChi) -> int:
            Get the value of the given board from the tablebase.

        value_white(board: boards.BoardChi) -> int:
            Get the value of the given board for the white player.

        get_over_tag(board: boards.BoardChi) -> OverTags:
            Get the over tag for the given board.

        string_result(board: boards.BoardChi) -> str:
            Get the string representation of the result for the given board.

        dtz(board: boards.BoardChi) -> int:
            Get the distance-to-zero (DTZ) value for the given board.

        best_move(board: boards.BoardChi) -> chess.Move:
            Get the best move according to the tablebase for the given board.
    """

    table_base: chess.syzygy.Tablebase

    def __init__(self, path_to_table: path):
        """
        Initialize the SyzygyTable object.

        Args:
            path_to_table (path): The path to the Syzygy tablebase.
        """
        path_to_table_str: str = str(path_to_table)
        self.table_base = chess.syzygy.open_tablebase(directory=path_to_table_str)

    def wdl(self, board: boards.BoardChi) -> int:
        return self.table_base.probe_wdl(board=board.chess_board)

    def dtz(self, board: boards.BoardChi) -> int:
        """
        Get the distance-to-zero (DTZ) value for the given board.

        Args:
            board (boards.BoardChi): The board to get the DTZ value for.

        Returns:
            int: The DTZ value for the board.
        """
        dtz: int = self.table_base.probe_dtz(board.chess_board)
        return dtz



"""
Module for creating player observers.
"""

import multiprocessing
import queue
from functools import partial
from typing import Protocol

import chess

from chipiron.utils import Seed
from chipiron.utils.communication.player_game_messages import BoardMessage
from chipiron.utils.dataclass import IsDataclass

from ..environments.chess_env.board.utils import FenPlusHistory
from ..scripts.chipiron_args import ImplementationArgs
from .boardevaluators.table_base.factory import AnySyzygyTable
from .factory import create_game_player
from .game_player import (
    GamePlayer,
    game_player_computes_move_on_board_and_send_move_in_queue,
)
from .player_args import PlayerFactoryArgs
from .player_thread import PlayerProcess


# function that will be called by the observable game when the board is updated, which should query at least one player
# to compute a move
# MoveFunction = Callable[[BoardChi, seed], None]
class MoveFunction(Protocol):
    """
    Represents a move function that can be called on a game board.

    Args:
        board (BoardChi): The game board on which the move function is applied.
        seed_int (seed): The seed used for the move function.

    Returns:
        None: This function does not return any value.
    """

    def __call__(self, fen_plus_history: FenPlusHistory, seed_int: Seed) -> None: ...


def send_board_to_player_process_mailbox(
    fen_plus_history: FenPlusHistory,
    seed_int: int,
    player_process_mailbox: queue.Queue[BoardMessage],
) -> None:
    """Sends the board and seed to the player process mailbox.

    This function creates a BoardMessage object with the given board and seed,
    and puts it into the player_process_mailbox.

    Args:
        seed_int (int): The seed to send.
        player_process_mailbox (queue.Queue[BoardMessage]): The mailbox to put the message into.
    """
    message: BoardMessage = BoardMessage(fen_plus_moves=fen_plus_history, seed=seed_int)
    player_process_mailbox.put(item=message)


class PlayerObserverFactory(Protocol):
    """
    creates the player and the means to communicate with it
    """

    def __call__(
        self,
        player_factory_args: PlayerFactoryArgs,
        player_color: chess.Color,
        main_thread_mailbox: queue.Queue[IsDataclass],
    ) -> tuple[GamePlayer | PlayerProcess, MoveFunction]: ...


def create_player_observer_factory(
    each_player_has_its_own_thread: bool,
    implementation_args: ImplementationArgs,
    universal_behavior: bool,
    syzygy_table: AnySyzygyTable | None,
) -> PlayerObserverFactory:
    """Create a player observer factory.
    This function creates a player observer factory based on the given parameters.
    The factory can create player observers that either run in separate threads or in the same process.
    The choice is determined by the `each_player_has_its_own_thread` parameter.
    The factory also takes into account the `implementation_args` and `universal_behavior` parameters
    to customize the behavior of the created player observers.
    The `syzygy_table` parameter is used to provide a syzygy table for the player observers.
    The factory returns a callable that can be used to create player observers with the specified arguments.
    The created player observers can be used to interact with the game and make moves.
    The `each_player_has_its_own_thread` parameter determines whether each player runs in its own thread
    or in the same process.
    The `implementation_args` parameter provides additional arguments for the implementation of the player observers.
    The `universal_behavior` parameter determines whether the player observers should exhibit universal behavior.

    Args:
        each_player_has_its_own_thread (bool): _description_
        implementation_args (ImplementationArgs): _description_
        universal_behavior (bool): _description_
        syzygy_table (AnySyzygyTable | None): _description_

    Returns:
        PlayerObserverFactory: _description_
    """
    player_observer_factory: PlayerObserverFactory
    if each_player_has_its_own_thread:
        player_observer_factory = partial(
            create_player_observer_distributed_players,
            implementation_args=implementation_args,
            universal_behavior=universal_behavior,
        )
    else:
        player_observer_factory = partial(
            create_player_observer_mono_process,
            syzygy_table=syzygy_table,
            universal_behavior=universal_behavior,
            implementation_args=implementation_args,
        )
    return player_observer_factory


def create_player_observer_distributed_players(
    player_factory_args: PlayerFactoryArgs,
    player_color: chess.Color,
    main_thread_mailbox: queue.Queue[IsDataclass],
    implementation_args: ImplementationArgs,
    universal_behavior: bool,
) -> tuple[GamePlayer | PlayerProcess, MoveFunction]:
    """Create a player observer.

    This function creates a player observer based on the given parameters.

    Args:
        board_factory: the board factory to create a board
        player_factory_args (PlayerFactoryArgs): The arguments for creating the player.
        player_color (chess.Color): The color of the player.
        main_thread_mailbox (queue.Queue[IsDataclass]): The mailbox for communication between the main thread
        and the player.

    Returns:
        tuple[ PlayerProcess, MoveFunction]: A tuple containing the player observer and the move function.

    """
    generic_player: PlayerProcess
    move_function: MoveFunction

    # case with multiprocessing when each player is a separate process
    # creating objects Queue that is the mailbox for the player thread
    player_process_mailbox = multiprocessing.Manager().Queue()

    # creating and starting the thread for the player
    player_process: PlayerProcess = PlayerProcess(
        player_factory_args=player_factory_args,
        player_color=player_color,
        queue_receiving_board=player_process_mailbox,
        queue_sending_move=main_thread_mailbox,
        implementation_args=implementation_args,
        universal_behavior=universal_behavior,
    )
    player_process.start()
    generic_player = player_process

    move_function = partial(
        send_board_to_player_process_mailbox,
        player_process_mailbox=player_process_mailbox,
    )

    return generic_player, move_function


def create_player_observer_mono_process(
    player_factory_args: PlayerFactoryArgs,
    player_color: chess.Color,
    main_thread_mailbox: queue.Queue[IsDataclass],
    syzygy_table: AnySyzygyTable | None,
    implementation_args: ImplementationArgs,
    universal_behavior: bool,
) -> tuple[GamePlayer, MoveFunction]:
    """Create a player observer.

    This function creates a player observer based on the given parameters.

    Args:
        player_factory_args (PlayerFactoryArgs): The arguments for creating the player.
        player_color (chess.Color): The color of the player.
        main_thread_mailbox (queue.Queue[IsDataclass]): The mailbox for communication between the main thread
        and the player.

    Returns:
        tuple[GamePlayer | PlayerProcess, MoveFunction]: A tuple containing the player observer and the move function.

    """
    generic_player: GamePlayer
    move_function: MoveFunction

    # case without multiprocessing all players and match manager in the same process

    generic_player = create_game_player(
        player_factory_args=player_factory_args,
        player_color=player_color,
        syzygy_table=syzygy_table,
        queue_progress_player=main_thread_mailbox,
        implementation_args=implementation_args,
        universal_behavior=universal_behavior,
    )
    move_function = partial(
        game_player_computes_move_on_board_and_send_move_in_queue,
        game_player=generic_player,
        queue_move=main_thread_mailbox,
    )

    return generic_player, move_function



"""
This module defines the `MatchArgs` class, which represents the input arguments needed by the one match script to run.
"""

from dataclasses import dataclass

from chipiron.games.match.match_settings_args import MatchSettingsArgs
from chipiron.games.match.match_tag import MatchConfigTag
from chipiron.players import PlayerArgs
from chipiron.players.player_ids import PlayerConfigTag


@dataclass
class MatchArgs:
    """
    The input arguments needed by the one match script to run

            file_name_player_one (path): The file name for player one. Defaults to 'RecurZipfBase3.yaml'.
            file_name_player_two (path): The file name for player two. Defaults to 'RecurZipfBase3.yaml'.
            file_name_match_setting (path): The file name for the match setting. Defaults to 'setting_cubo.yaml'.
            player_one (dict[Any, Any]): The dictionary for player one. Defaults to an empty dictionary.
            player_two (dict[Any, Any]): The dictionary for player two. Defaults to an empty dictionary.
            match (dict[Any, Any]): The dictionary for the match. Defaults to an empty dictionary.

    """

    player_one: PlayerConfigTag | PlayerArgs = PlayerConfigTag.RECUR_ZIPF_BASE_3
    player_two: PlayerConfigTag | PlayerArgs = PlayerConfigTag.RECUR_ZIPF_BASE_3
    match_setting: MatchConfigTag | MatchSettingsArgs = MatchConfigTag.CUBO



"""
Module for neural network board evaluators.
"""

from .neural_net_board_eval_args import NeuralNetBoardEvalArgs
from .nn_board_evaluator import NNBoardEvaluator

__all__ = ["NNBoardEvaluator", "NeuralNetBoardEvalArgs"]



"""
This module defines the PlayerConfigTag enumeration, which identifies different player configuration files for the Chipiron chess engine framework. Each tag corresponds to a specific player type, including human interfaces, AI agents, and external engines. The PlayerConfigTag class provides utility methods to determine if a player is human, retrieve the YAML configuration file path for a player, and load player arguments from the configuration file.

Classes:
    PlayerConfigTag (Enum): Enumeration of player configuration tags with methods for player type checking and configuration retrieval.

Methods:
    is_human(): Returns True if the player configuration tag represents a human player.
    get_yaml_file_path(): Returns the file path to the YAML configuration file for the player.
    get_players_args(): Loads and returns player arguments from the corresponding YAML configuration file.

Dependencies:
    - enum
    - importlib.resources
    - parsley_coco
    - chipiron.players.PlayerArgs
    - chipiron.utils.path

"""

from enum import Enum
from importlib.resources import as_file, files

import parsley_coco

from chipiron.players import PlayerArgs
from chipiron.utils import path


class PlayerConfigTag(str, Enum):
    """
    This class is used to identify the player configuration files.
    Each player configuration file should be listed here as a class attribute.
    The class also provides methods to check if a player is human and to get the YAML file path for the player configuration.
    """

    CHIPIRON = "Chipiron"

    GUI_HUMAN = "Gui_Human"
    COMMAND_LINE_HUMAN = "Command_Line_Human"
    SEQUOOL = "Sequool"
    RECUR_ZIPF_BASE_3 = "RecurZipfBase3"
    RECUR_ZIPF_BASE_4 = "RecurZipfBase4"
    UNIFORM_DEPTH = "UniformDepth"
    UNIFORM_DEPTH_2 = "UniformDepth2"
    UNIFORM_DEPTH_3 = "UniformDepth3"
    UNIFORM = "Uniform"
    RANDOM = "Random"
    STOCKFISH = "Stockfish"

    def is_human(self) -> bool:
        """Check if the player is human.
        This method checks if the player is a human player based on the player's configuration tag.
        It returns True if the player is a human player, and False otherwise.

        Returns:
            bool: True if the player is a human player, False otherwise.
        """
        return (
            self is PlayerConfigTag.GUI_HUMAN
            or self is PlayerConfigTag.COMMAND_LINE_HUMAN
        )

    def get_yaml_file_path(self) -> path:
        """Get the YAML file path for the player configuration.
        This method constructs the file path for the player configuration YAML file
        based on the player's configuration tag.
        It returns the file path as a string.

        Returns:
            path: The file path for the player configuration YAML file.
        Raises:
            ValueError: If the player configuration tag is not recognized.
        """
        if self is PlayerConfigTag.CHIPIRON:
            resource = files("chipiron").joinpath(
                "data/players/player_config/chipiron/chipiron.yaml"
            )
        else:
            resource = files("chipiron").joinpath(
                f"data/players/player_config/{self.value}.yaml"
            )

        with as_file(resource) as real_path:
            return str(real_path)  # Or pass it immediately to another function

    def get_players_args(self) -> PlayerArgs:
        """Get the player arguments from the YAML file.
        This method fetches the player arguments from the YAML file
        corresponding to the player's configuration tag.

        Returns:
            PlayerArgs: The player arguments as a dataclass.
        """
        player_args: PlayerArgs = parsley_coco.resolve_yaml_file_to_base_dataclass(
            yaml_path=str(self.get_yaml_file_path()),
            base_cls=PlayerArgs,
            raise_error_with_nones=False,
            package_name=str(files("chipiron")),
        )
        return player_args


if __name__ == "__main__":
    a = PlayerConfigTag.COMMAND_LINE_HUMAN



"""
Evaluates neural network models on a test dataset and records their performance.
Saves the evaluation results to a YAML file.
"""

import datetime
import os
import time
from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Any

import dacite
import torch
import yaml
from torch.utils.data import DataLoader

from chipiron.learningprocesses.nn_trainer.nn_trainer import (
    compute_test_error_on_dataset,
)
from chipiron.players.boardevaluators.datasets.datasets import (
    FenAndValueData,
    FenAndValueDataSet,
    custom_collate_fn_fen_and_value,
    process_stockfish_value,  # pyright: ignore[reportUnknownVariableType]
)
from chipiron.players.boardevaluators.neural_networks.factory import (
    NeuralNetModelsAndArchitecture,
    create_nn_board_eval_from_nn_parameters_file_and_existing_model,
)
from chipiron.utils import path
from chipiron.utils.chi_nn import ChiNN

if TYPE_CHECKING:
    from chipiron.players.boardevaluators.neural_networks import NNBoardEvaluator


@dataclass
class ModelEvaluation:
    """
    Represents a model evaluation.
    """

    evaluation: float
    time_of_evaluation: datetime.datetime
    number_of_model_parameters: int


type EvaluatedModels = dict[str, ModelEvaluation]


def compute_model_hash_key(model_and_archi: NeuralNetModelsAndArchitecture) -> str:
    """Computes a unique hash key for the model and architecture.

    Args:
        model_and_archi (NeuralNetModelsAndArchitecture): The model and architecture to compute the hash key for.

    Returns:
        str: The computed hash key.
    """
    return (
        str(model_and_archi.model_weights_file_name)
        + model_and_archi.nn_architecture_args.filename()
    )


def count_parameters(model: ChiNN) -> int:
    """Counts the number of trainable parameters in a model."""
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def evaluate_models(
    models_to_evaluate: list[NeuralNetModelsAndArchitecture],
    evaluation_report_file: path = "chipiron/scripts/evaluate_models/evaluation_report.yaml",
    dataset_file_name: path = "external_data/data_chipiron/datasets/goodgames_plusvariation_stockfish_eval_test",
) -> None:
    """
    Evaluates the models in the list models_to_evaluate.

    Raises:
        Exception: If there is an error loading the evaluation report.
    """
    print("Evaluating models...")
    # Load the evaluation report
    evaluated_models: dict[str, ModelEvaluation] = {}
    with open(evaluation_report_file, "r", encoding="utf-8") as stream:
        try:
            evaluated_models_temp: dict[str, dict[Any, Any]] | None
            evaluated_models_temp = yaml.safe_load(stream)
            if evaluated_models_temp is None:
                evaluated_models_temp = {}
            evaluated_model_path: path
            evaluated_model_evaluation_dict: dict[Any, Any]
            for (
                evaluated_model_path,
                evaluated_model_evaluation_dict,
            ) in evaluated_models_temp.items():
                model_evaluation_: ModelEvaluation = dacite.from_dict(
                    data_class=ModelEvaluation, data=evaluated_model_evaluation_dict
                )
                evaluated_models[evaluated_model_path] = model_evaluation_
            print(
                "Here is the content of the previous evaluation report: ",
                evaluated_models,
            )
        except yaml.YAMLError as exc:
            print(exc)
            raise yaml.YAMLError("Error loading the evaluation report") from exc

    # Evaluate the models if necessary
    data_loader_stockfish_boards_test = None
    model_to_evaluate: NeuralNetModelsAndArchitecture
    for model_to_evaluate in models_to_evaluate:
        print(f"\nEvaluating model: {model_to_evaluate}")

        model_hash_key: str = compute_model_hash_key(model_to_evaluate)

        # Check if the model has already been evaluated
        should_evaluate: bool
        if model_hash_key in evaluated_models:
            model_evaluation: ModelEvaluation = evaluated_models[model_hash_key]
            print(
                "This model has already been evaluated. Now checking if the file has been modified since last evaluation..."
            )
            model_params_path_pt: path
            model_params_path_pt = model_to_evaluate.model_weights_file_name

            modification_time = os.path.getmtime(model_params_path_pt)
            readable_time = time.ctime(modification_time)
            print(
                f"Last modification time: {readable_time} and last evaluation time {model_evaluation.time_of_evaluation} "
            )
            # Check if the model has been modified since last evaluation
            if modification_time > model_evaluation.time_of_evaluation.timestamp():
                # The model has been modified since last evaluation
                print(
                    "The model has been modified since last evaluation. Asking for evaluation..."
                )
                should_evaluate = True
            else:
                # The model has not been modified since last evaluation
                print(
                    "The model has not been modified since last evaluation. No need to evaluate."
                )
                should_evaluate = False
        else:
            print("This model has not been evaluated yet. Asking for evaluation...")
            should_evaluate = True

        # Check if the model should be evaluated
        if should_evaluate:
            # Evaluate the model
            print("Evaluating model...")

            criterion = torch.nn.L1Loss()

            nn_board_evaluator: NNBoardEvaluator
            nn_board_evaluator = (
                create_nn_board_eval_from_nn_parameters_file_and_existing_model(
                    model_weights_file_name=model_to_evaluate.model_weights_file_name,
                    nn_architecture_args=model_to_evaluate.nn_architecture_args,
                )
            )

            stockfish_boards_test = FenAndValueDataSet(
                file_name=dataset_file_name,
                preprocessing=False,
                transform_board_function=nn_board_evaluator.board_to_input_convert,
                transform_dataset_value_to_white_value_function=process_stockfish_value,  # pyright: ignore[reportUnknownArgumentType]
                transform_white_value_to_model_output_function=nn_board_evaluator.output_and_value_converter.from_value_white_to_model_output,
            )

            stockfish_boards_test.load()

            data_loader_stockfish_boards_test = DataLoader[FenAndValueData](
                stockfish_boards_test,
                batch_size=50000,
                shuffle=False,
                num_workers=1,
                collate_fn=custom_collate_fn_fen_and_value,
            )
            print(f"Size of test set: {len(data_loader_stockfish_boards_test)}")

            evalu = compute_test_error_on_dataset(
                net=nn_board_evaluator.net,
                criterion=criterion,
                data_test=data_loader_stockfish_boards_test,
                number_of_tests=len(data_loader_stockfish_boards_test),
            )
            number_of_model_parameters: int = count_parameters(nn_board_evaluator.net)
            model_evaluation = ModelEvaluation(
                evaluation=evalu,
                time_of_evaluation=datetime.datetime.now(),
                number_of_model_parameters=number_of_model_parameters,
            )
            evaluated_models[model_hash_key] = model_evaluation
            print("Model evaluated!")

    evaluated_model_path_: path
    evaluated_model_evaluation: ModelEvaluation
    evaluated_models_final_dict: dict[path, dict[Any, Any]] = {}
    for evaluated_model_path_, evaluated_model_evaluation in evaluated_models.items():
        evaluated_models_final_dict[evaluated_model_path_] = asdict(
            evaluated_model_evaluation
        )
    with open(
        "chipiron/scripts/evaluate_models/evaluation_report.yaml", "w", encoding="utf-8"
    ) as outfile:
        yaml.dump(evaluated_models_final_dict, outfile, default_flow_style=False)

    print("All evaluations done!")


if __name__ == "__main__":
    models_to_evaluate_: list[NeuralNetModelsAndArchitecture] = [
        NeuralNetModelsAndArchitecture.build_from_folder_path(
            folder_path="chipiron/scripts/learn_nn_supervised/board_evaluators_common_training_data/nn_pytorch/prelu_no_bug"
        ),
        NeuralNetModelsAndArchitecture.build_from_folder_path(
            folder_path="chipiron/scripts/learn_nn_supervised/board_evaluators_common_training_data/nn_pytorch/nn_p1_new"
        ),
        NeuralNetModelsAndArchitecture.build_from_folder_path(
            folder_path="chipiron/scripts/learn_nn_supervised/board_evaluators_common_training_data/nn_pytorch/nn_pp2_new"
        ),
        NeuralNetModelsAndArchitecture.build_from_folder_path(
            folder_path="chipiron/scripts/learn_nn_supervised/board_evaluators_common_training_data/nn_pytorch/nn_pp2d3_prelu_new"
        ),
        # NeuralNetModelsAndArchitecture.build_from_folder_path(
        #     folder_path="chipiron/scripts/learn_nn_supervised/board_evaluators_common_training_data/nn_pytorch/transformerone"),
    ]

    evaluate_models(models_to_evaluate=models_to_evaluate_)



"""
Module to transform the board into a tensor representation.
"""

import chess
import torch

from chipiron.environments.chess_env.board import IBoard


def transform_board_pieces_one_side(
    board: IBoard, requires_grad_bool: bool
) -> torch.Tensor:
    """
    Transform the board pieces for one side into a tensor representation.

    Args:
        board (BoardChi): The chess board.
        requires_grad_bool (bool): Whether the tensor requires gradient.

    Returns:
        torch.Tensor: The transformed board pieces tensor.
    """
    # normalisation of the board so that it is white turn (possible color inversion if it was black's turn)

    if board.turn == chess.BLACK:
        color_turn = board.turn
        color_not_turn = chess.WHITE
    else:
        color_turn = chess.WHITE
        color_not_turn = chess.BLACK

    transform = torch.zeros(5)

    transform[0] = bin(board.pawns & board.occupied_color(color_turn)).count("1") - bin(
        board.pawns & board.occupied_color(color_not_turn)
    ).count("1")
    transform[1] = bin(board.knights & board.occupied_color(color_turn)).count(
        "1"
    ) - bin(board.knights & board.occupied_color(color_not_turn)).count("1")
    transform[2] = bin(board.bishops & board.occupied_color(color_turn)).count(
        "1"
    ) - bin(board.bishops & board.occupied_color(color_not_turn)).count("1")
    transform[3] = bin(board.rooks & board.occupied_color(color_turn)).count("1") - bin(
        board.rooks & board.occupied_color(color_not_turn)
    ).count("1")
    transform[4] = bin(board.queens & board.occupied_color(color_turn)).count(
        "1"
    ) - bin(board.queens & board.occupied_color(color_not_turn)).count("1")

    if requires_grad_bool:
        transform.requires_grad_(True)

    return transform


def transform_board_pieces_two_sides(
    board: IBoard, requires_grad_bool: bool
) -> torch.Tensor:
    """
    Transform the board pieces for both sides into a tensor representation.

    Args:
        board (BoardChi): The chess board.
        requires_grad_bool (bool): Whether the tensor requires gradient.

    Returns:
        torch.Tensor: The transformed board pieces tensor.
    """
    # normalisation of the board so that it is white turn (possible color inversion if it was black's turn)

    if board.turn == chess.BLACK:
        color_turn = board.turn
        color_not_turn = chess.WHITE
    else:
        color_turn = chess.WHITE
        color_not_turn = chess.BLACK

    transform = torch.zeros(10, requires_grad=requires_grad_bool)

    transform[0] = bin(board.pawns & board.occupied_color(color_turn)).count("1")
    transform[1] = bin(board.knights & board.occupied_color(color_turn)).count("1")
    transform[2] = bin(board.bishops & board.occupied_color(color_turn)).count("1")
    transform[3] = bin(board.rooks & board.occupied_color(color_turn)).count("1")
    transform[4] = bin(board.queens & board.occupied_color(color_turn)).count("1")
    transform[5] = -bin(board.pawns & board.occupied_color(color_not_turn)).count("1")
    transform[6] = -bin(board.knights & board.occupied_color(color_not_turn)).count("1")
    transform[7] = -bin(board.bishops & board.occupied_color(color_not_turn)).count("1")
    transform[8] = -bin(board.rooks & board.occupied_color(color_not_turn)).count("1")
    transform[9] = -bin(board.queens & board.occupied_color(color_not_turn)).count("1")
    return transform


def get_tensor_from_tensors(
    tensor_white: torch.Tensor,
    tensor_black: torch.Tensor,
    tensor_castling_white: torch.Tensor,
    tensor_castling_black: torch.Tensor,
    color_to_play: chess.Color,
) -> torch.Tensor:
    """
    Get the final tensor representation from individual tensors.

    Args:
        tensor_white (torch.Tensor): The tensor representation for white pieces.
        tensor_black (torch.Tensor): The tensor representation for black pieces.
        tensor_castling_white (torch.Tensor): The tensor representation for white castling.
        tensor_castling_black (torch.Tensor): The tensor representation for black castling.
        color_to_play (chess.Color): The color to play.

    Returns:
        torch.Tensor: The final tensor representation.
    """
    if color_to_play == chess.WHITE:
        tensor = tensor_white - tensor_black
    else:
        tensor = tensor_black - tensor_white

    if color_to_play == chess.WHITE:
        tensor_castling = tensor_castling_white - tensor_castling_black
    else:
        tensor_castling = tensor_castling_black - tensor_castling_white

    tensor_2 = torch.cat((tensor, tensor_castling), 0)
    return tensor_2



"""
This module defines the Comparable class, which is an abstract base class for objects that can be compared.

Classes that inherit from Comparable must implement the __lt__ method, which defines the less-than comparison operation.

"""

from abc import ABCMeta, abstractmethod
from typing import Any, TypeVar


class Comparable(metaclass=ABCMeta):
    """
    An abstract base class for objects that can be compared.

    Subclasses of `Comparable` must implement the `__lt__` method to define the less-than comparison.

    Attributes:
        None

    Methods:
        __lt__(self, other: Any) -> bool: Abstract method that compares the object with another object and
         returns True if the object is less than the other object, False otherwise.
    """

    @abstractmethod
    def __lt__(self, other: Any) -> bool: ...


CT = TypeVar("CT", bound=Comparable)



"""
This module defines the NodeSelectorType enumeration, which represents the types of node selectors.
"""

from enum import Enum


class NodeSelectorType(str, Enum):
    """
    Enumeration representing the types of node selectors.
    """

    RECUR_ZIPF_BASE = "RecurZipfBase"
    SEQUOOL = "Sequool"
    UNIFORM = "Uniform"



"""
Module that contains small tools that are used in the project.

"""

from .small_tools import (
    Seed,
    dict_alphabetic_str,
    mkdir_if_not_existing,
    path,
    rec_merge_dic,
    unique_int_from_list,
    yaml_fetch_args_in_file,
)

__all__ = [
    "yaml_fetch_args_in_file",
    "rec_merge_dic",
    "Seed",
    "path",
    "mkdir_if_not_existing",
    "unique_int_from_list",
    "dict_alphabetic_str",
]



"""
Module for creating neural networks and neural network board evaluators.
"""

import os.path
import sys
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any

import dacite

from chipiron.players.boardevaluators.neural_networks.input_converters.board_to_input import (
    BoardToInputFunction,
    create_board_to_input,
)
from chipiron.players.boardevaluators.neural_networks.models.multi_layer_perceptron import (
    MultiLayerPerceptron,
    MultiLayerPerceptronArgs,
)
from chipiron.players.boardevaluators.neural_networks.models.transformer_one import (
    TransformerArgs,
    TransformerOne,
)
from chipiron.players.boardevaluators.neural_networks.neural_net_architecture_args import (
    NeuralNetArchitectureArgs,
)
from chipiron.players.boardevaluators.neural_networks.nn_board_evaluator import (
    NNBoardEvaluator,
)
from chipiron.players.boardevaluators.neural_networks.NNModelTypeArgs import (
    NNModelTypeArgs,
)
from chipiron.players.boardevaluators.neural_networks.output_converters.factory import (
    create_output_converter,
)
from chipiron.utils import path, yaml_fetch_args_in_file
from chipiron.utils.chi_nn import ChiNN

if TYPE_CHECKING:
    from chipiron.players.boardevaluators.neural_networks.output_converters.output_value_converter import (
        OutputValueConverter,
    )


@dataclass
class NeuralNetModelsAndArchitecture:
    """
    Class to hold the neural network models and architecture.
    Attributes:
        model_weights_file_name (path): The file name of the model weights.
        nn_architecture_args (NeuralNetArchitectureArgs): The neural network architecture arguments.
    """

    model_weights_file_name: path
    nn_architecture_args: NeuralNetArchitectureArgs

    @classmethod
    def build_from_folder_path(
        cls, folder_path: path
    ) -> "NeuralNetModelsAndArchitecture":
        """
        Build an instance of NeuralNetModelsAndArchitecture from the given folder path.

        Args:
            folder_path (Path): Path to the folder containing 'architecture.yaml' and model weights.

        Returns:
            NeuralNetModelsAndArchitecture: An initialized instance.
        """
        nn_args = get_architecture_args_from_folder(folder_path=folder_path)
        model_file = os.path.join(folder_path, nn_args.filename() + ".pt")

        return cls(model_weights_file_name=model_file, nn_architecture_args=nn_args)


def get_nn_param_file_path_from(
    folder_path: path, file_name: str | None = None
) -> tuple[str, str]:
    """
    Get the file path for the neural network parameters.

    Args:
        folder_path (str): The folder path for the neural network parameters.

    Returns:
        str: The file path for the neural network parameters.
    """
    nn_param_file_path: str
    if file_name is None:
        nn_param_file_path = os.path.join(folder_path, "param")
    else:
        nn_param_file_path = os.path.join(folder_path, file_name)
    return nn_param_file_path + ".pt", nn_param_file_path + ".yaml"


def get_nn_architecture_file_path_from(folder_path: path) -> str:
    """
    Get the file path for the architecture parameters.

    Args:
        folder_path (str): The folder path for the architecture parameters.

    Returns:
        str: The file path for the architecture parameters.
    """
    nn_param_file_path: str = os.path.join(folder_path, "architecture.yaml")
    return nn_param_file_path


def create_nn(nn_type_args: NNModelTypeArgs) -> ChiNN:
    """
    Create a neural network.
    """

    net: ChiNN
    match nn_type_args:
        case MultiLayerPerceptronArgs():
            net = MultiLayerPerceptron(args=nn_type_args)
        case TransformerArgs():
            net = TransformerOne(args=nn_type_args)
        case _:
            sys.exit(f"Create NN: can not find {nn_type_args} in file {__name__}")
    return net


# todo: probably dead code, check!
def get_architecture_args_from_file(
    architecture_file_name: path,
) -> NeuralNetArchitectureArgs:
    """Get the architecture arguments from a YAML file.

    Args:
        architecture_file_name (path): The path to the architecture YAML file.

    Returns:
        NeuralNetArchitectureArgs: The architecture arguments.
    """
    args_dict: dict[Any, Any] = yaml_fetch_args_in_file(
        path_file=architecture_file_name
    )
    nn_architecture_args: NeuralNetArchitectureArgs = dacite.from_dict(
        data_class=NeuralNetArchitectureArgs,
        data=args_dict,
        config=dacite.Config(cast=[Enum]),
    )
    return nn_architecture_args


def get_architecture_args_from_folder(folder_path: path) -> NeuralNetArchitectureArgs:
    """Get the architecture arguments from a folder.
    Args:
        folder_path (path): The path to the folder containing the architecture file.
    Returns:
        NeuralNetArchitectureArgs: The architecture arguments.
    """
    architecture_file_name: path = get_nn_architecture_file_path_from(
        folder_path=folder_path
    )
    if not os.path.isfile(architecture_file_name):
        raise FileNotFoundError(f"this is not a file {architecture_file_name}")

    nn_architecture_args: NeuralNetArchitectureArgs = get_architecture_args_from_file(
        architecture_file_name=architecture_file_name
    )

    return nn_architecture_args


def create_nn_from_param_path_and_architecture_args(
    model_weights_file_name: path, nn_architecture_args: NeuralNetArchitectureArgs
) -> tuple[ChiNN, NeuralNetArchitectureArgs]:
    """Create a neural network from a parameter file and architecture arguments.
    Args:
        model_weights_file_name (path): The path to the model weights file.
        nn_architecture_args (NeuralNetArchitectureArgs): The architecture arguments.
    Returns:
        tuple[ChiNN, NeuralNetArchitectureArgs]: The created neural network and the architecture arguments.
    """
    net: ChiNN = create_nn(nn_type_args=nn_architecture_args.model_type_args)
    net.load_weights_from_file(path_to_param_file=model_weights_file_name)
    return net, nn_architecture_args


def create_nn_from_folder_path_and_existing_model(
    folder_path: path,
) -> tuple[ChiNN, NeuralNetArchitectureArgs]:
    """Create a neural network from a folder path and existing model.

    Args:
        folder_path (path): The path to the folder containing the model files.

    Returns:
        tuple[ChiNN, NeuralNetArchitectureArgs]: _description_
    """
    nn_architecture_args: NeuralNetArchitectureArgs = get_architecture_args_from_folder(
        folder_path=folder_path
    )
    model_weights_file_name: path = os.path.join(folder_path, "param.pt")

    return create_nn_from_param_path_and_architecture_args(
        model_weights_file_name=model_weights_file_name,
        nn_architecture_args=nn_architecture_args,
    )


def create_nn_board_eval_from_folder_path_and_existing_model(
    path_to_nn_folder: path,
) -> tuple[NNBoardEvaluator, NeuralNetArchitectureArgs]:
    """
    Create a neural network board evaluator.

    Args:
        path_to_nn_folder (path): the path to the folder where the model is defined.

    Returns:
        NNBoardEvaluator: The created neural network board evaluator.
    """
    net: ChiNN
    nn_architecture_args: NeuralNetArchitectureArgs
    net, nn_architecture_args = create_nn_from_folder_path_and_existing_model(
        folder_path=path_to_nn_folder
    )

    nn_board_evaluator = create_nn_board_eval_from_nn_and_architecture_args(
        nn=net, nn_architecture_args=nn_architecture_args
    )
    return nn_board_evaluator, nn_architecture_args


def create_nn_board_eval_from_nn_and_architecture_args(
    nn_architecture_args: NeuralNetArchitectureArgs,
    nn: ChiNN,
) -> NNBoardEvaluator:
    board_to_input_convert: BoardToInputFunction = create_board_to_input(
        model_input_representation_type=nn_architecture_args.model_input_representation_type
    )

    output_and_value_converter: OutputValueConverter = create_output_converter(
        model_output_type=nn_architecture_args.model_output_type
    )

    return NNBoardEvaluator(
        net=nn,
        output_and_value_converter=output_and_value_converter,
        board_to_input_convert=board_to_input_convert,
    )


def create_nn_board_eval_from_architecture_args(
    nn_architecture_args: NeuralNetArchitectureArgs,
) -> NNBoardEvaluator:
    nn = create_nn(nn_type_args=nn_architecture_args.model_type_args)
    nn.init_weights()

    return create_nn_board_eval_from_nn_and_architecture_args(
        nn_architecture_args=nn_architecture_args, nn=nn
    )


def create_nn_board_eval_from_nn_parameters_file_and_existing_model(
    model_weights_file_name: path, nn_architecture_args: NeuralNetArchitectureArgs
) -> NNBoardEvaluator:
    net: ChiNN
    net, nn_architecture_args = create_nn_from_param_path_and_architecture_args(
        model_weights_file_name=model_weights_file_name,
        nn_architecture_args=nn_architecture_args,
    )

    return create_nn_board_eval_from_nn_and_architecture_args(
        nn_architecture_args=nn_architecture_args, nn=net
    )



"""
Module defining the board representation interface and the 364 features board representation.
"""

from dataclasses import dataclass
from typing import Protocol

import chess
import torch


class BoardRepresentation(Protocol):
    """
    Protocol defining the interface for a board representation.
    """

    def get_evaluator_input(self, color_to_play: chess.Color) -> torch.Tensor:
        """
        Returns the evaluator input tensor for the given color to play.

        Args:
            color_to_play: The color to play, either chess.WHITE or chess.BLACK.

        Returns:
            The evaluator input tensor.
        """
        ...


@dataclass(slots=True)
class Representation364(BoardRepresentation):
    """
    Dataclass representing a board representation with 364 features.
    """

    tensor_white: torch.Tensor
    tensor_black: torch.Tensor
    tensor_castling_white: torch.Tensor
    tensor_castling_black: torch.Tensor

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Representation364):
            return False
        else:
            return (
                torch.equal(self.tensor_white, other.tensor_white)
                and torch.equal(self.tensor_black, other.tensor_black)
                and torch.equal(self.tensor_castling_black, other.tensor_castling_black)
                and torch.equal(self.tensor_castling_white, other.tensor_castling_white)
            )

    def get_evaluator_input(self, color_to_play: chess.Color) -> torch.Tensor:
        """
        Returns the evaluator input tensor for the given color to play.

        Args:
            color_to_play: The color to play, either chess.WHITE or chess.BLACK.

        Returns:
            The evaluator input tensor.
        """
        if color_to_play == chess.WHITE:
            tensor = torch.cat((self.tensor_white, self.tensor_black), 0)
        else:
            tensor = torch.cat((self.tensor_black, self.tensor_white), 0)

        if color_to_play == chess.WHITE:
            tensor_castling = torch.cat(
                (self.tensor_castling_white, self.tensor_castling_black), 0
            )
        else:
            tensor_castling = torch.cat(
                (self.tensor_castling_black, self.tensor_castling_white), 0
            )

        tensor_2 = torch.cat((tensor, tensor_castling), 0)

        return tensor_2


@dataclass(slots=True)
class Representation364_2(BoardRepresentation):
    """
    Dataclass representing a board representation with 364 features.
    """

    tensor: torch.Tensor

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Representation364_2):
            return False
        else:
            return torch.equal(self.tensor, other.tensor)

    def get_evaluator_input(self, color_to_play: chess.Color) -> torch.Tensor:
        """
        Returns the evaluator input tensor for the given color to play.

        Args:
            color_to_play: The color to play, either chess.WHITE or chess.BLACK.

        Returns:
            The evaluator input tensor.
        """
        return self.tensor



"""
player_thread.py
"""

import multiprocessing
import queue
from typing import TYPE_CHECKING

import chess

from chipiron.environments.chess_env.board.factory import BoardFactory
from chipiron.utils.communication.player_game_messages import BoardMessage
from chipiron.utils.dataclass import DataClass, IsDataclass
from chipiron.utils.logger import chipiron_logger

from ..environments.chess_env.board.factory import create_board_factory
from ..scripts.chipiron_args import ImplementationArgs
from .boardevaluators.table_base.factory import SyzygyFactory, create_syzygy_factory
from .factory import create_game_player
from .game_player import (
    GamePlayer,
    game_player_computes_move_on_board_and_send_move_in_queue,
)
from .player_args import PlayerFactoryArgs

if TYPE_CHECKING:
    from chipiron.utils import Seed

    from ..environments.chess_env.board.utils import FenPlusHistory


# A class that extends the Thread class
class PlayerProcess(multiprocessing.Process):
    """A class representing a player process.

    This class extends the `multiprocessing.Process` class and is responsible for running a player
     in a separate process.

    Attributes:
        game_player (GamePlayer): The game player object.
        queue_board (queue.Queue[DataClass]): The queue for receiving board messages.
        queue_move (queue.Queue[IsDataclass]): The queue for sending move messages.
        player_color (chess.Color): The color of the player.

    Args:
        player_factory_args (PlayerFactoryArgs): The arguments for creating the game player.
        queue_board (queue.Queue[DataClass]): The queue for receiving board messages.
        queue_move (queue.Queue[IsDataclass]): The queue for sending move messages.
        player_color (chess.Color): The color of the player.
    """

    game_player: GamePlayer
    queue_receiving_board: queue.Queue[DataClass]
    queue_sending_move: queue.Queue[IsDataclass]
    player_color: chess.Color
    board_factory: BoardFactory

    def __init__(
        self,
        player_factory_args: PlayerFactoryArgs,
        queue_receiving_board: queue.Queue[DataClass],
        queue_sending_move: queue.Queue[IsDataclass],
        player_color: chess.Color,
        implementation_args: ImplementationArgs,
        universal_behavior: bool,
    ) -> None:
        """Initialize the PlayerThread object.

        Args:
            player_factory_args (PlayerFactoryArgs): The arguments required to create the player.
            queue_receiving_board (queue.Queue[DataClass]): The queue for receiving board data.
            queue_move (queue.Queue[IsDataclass]): The queue for sending move data.
            player_color (chess.Color): The color of the player.

        """
        # Call the Thread class's init function
        multiprocessing.Process.__init__(self, daemon=False)
        self._stop_event = multiprocessing.Event()
        self.queue_sending_move = queue_sending_move
        self.queue_receiving_board = queue_receiving_board
        self.player_color = player_color

        self.board_factory: BoardFactory = create_board_factory(
            use_rust_boards=implementation_args.use_rust_boards,
            use_board_modification=implementation_args.use_board_modification,
            sort_legal_moves=universal_behavior,
        )

        create_syzygy: SyzygyFactory = create_syzygy_factory(
            use_rust=implementation_args.use_rust_boards
        )

        self.game_player: GamePlayer = create_game_player(
            player_factory_args=player_factory_args,
            player_color=player_color,
            syzygy_table=create_syzygy(),
            queue_progress_player=queue_sending_move,
            implementation_args=implementation_args,
            universal_behavior=universal_behavior,
        )
        assert self.game_player.player is not None

    # Override the run() function of Thread class
    def run(self) -> None:
        """
        Executes the player thread.

        This method is called when the player thread is started. It continuously checks for messages in the message queue
        and handles them accordingly. If a message is a `BoardMessage`, it retrieves the board and seed from the message,
        computes the move for the board using the `game_player`, and sends the move to the move queue. If the message is
        not a `BoardMessage`, it simply prints the message.

        Note: This method runs indefinitely until the thread is stopped externally.

        Returns:
            None
        """
        chipiron_logger.info("Started player thread: %s", self.game_player)

        while True:
            try:
                message = self.queue_receiving_board.get(False)
            except queue.Empty:
                pass
            else:
                # Handle task here and call q.task_done()
                if isinstance(message, BoardMessage):
                    board_message: BoardMessage = message
                    fen_plus_moves: FenPlusHistory = board_message.fen_plus_moves
                    seed_: Seed | None = board_message.seed
                    chipiron_logger.info(
                        "Player thread got the board %s", fen_plus_moves.current_fen
                    )
                    assert seed_ is not None

                    # the game_player computes the move for the board and sends the move in the move queue
                    game_player_computes_move_on_board_and_send_move_in_queue(
                        fen_plus_history=fen_plus_moves,
                        game_player=self.game_player,
                        queue_move=self.queue_sending_move,
                        seed_int=seed_,
                    )
                else:
                    chipiron_logger.warning(
                        "NOT EXPECTING THIS MESSAGE !! : %s", message
                    )

            # TODO here give option to continue working while the other is thinking



"""
This module defines the ScriptType enum and provides a mapping between script types and their corresponding script classes.
"""

from enum import Enum


class ScriptType(str, Enum):
    """
    Enum class representing different types of scripts.
    """

    ONE_MATCH = "one_match"
    LEAGUE = "league"
    LEARN_NN = "learn_nn"
    LEARN_NN_FROM_SCRATCH = "learn_nn_from_scratch"
    BASE_TREE_EXPLORATION = "base_tree_exploration"
    TREE_VISUALIZATION = "tree_visualization"
    REPLAY_MATCH = "replay_match"



"""
Basic class for Creating Tree nodes
"""

import chipiron.environments.chess_env.board as board_mod
import chipiron.environments.chess_env.board as boards
from chipiron.environments.chess_env.move.imove import moveKey
from chipiron.players.move_selector.treevalue.node_factory.node_factory import (
    TreeNodeFactory,
)
from chipiron.players.move_selector.treevalue.nodes.itree_node import ITreeNode
from chipiron.players.move_selector.treevalue.nodes.tree_node import TreeNode


class Base[T: ITreeNode = ITreeNode](TreeNodeFactory[T]):
    """
    Basic class for Creating Tree nodes
    """

    def create(
        self,
        board: boards.IBoard,
        half_move: int,
        count: int,
        parent_node: ITreeNode | None,
        move_from_parent: moveKey | None,
        modifications: board_mod.BoardModificationP | None,
    ) -> TreeNode[T]:
        """
        Creates a new TreeNode object.

        Args:
            board (boards.BoardChi): The current board state.
            half_move (int): The half-move count.
            count (int): The ID of the new node.
            parent_node (ITreeNode | None): The parent node of the new node.
            move_from_parent (chess.Move | None): The move that leads to the new node.
            modifications (board_mod.BoardModification | None): The modifications applied to the board.

        Returns:
            TreeNode: The newly created TreeNode object.
        """

        parent_nodes: dict[ITreeNode, moveKey]
        if parent_node is None:
            parent_nodes = {}
        else:
            assert move_from_parent is not None
            parent_nodes = {parent_node: move_from_parent}

        tree_node: TreeNode[T] = TreeNode(
            board_=board,
            half_move_=half_move,
            id_=count,
            parent_nodes_=parent_nodes,
        )
        return tree_node



"""
Module containing the different messages that can be sent to the GUI.
"""

from .evaluation_messsage import EvaluationMessage
from .game_status_message import BackMessage, GameStatusMessage
from .progress_messsage import PlayerProgressMessage

__all__ = [
    "GameStatusMessage",
    "EvaluationMessage",
    "BackMessage",
    "PlayerProgressMessage",
]



import chess
import pytest

from chipiron.environments.chess_env.board import IBoard, create_board
from chipiron.environments.chess_env.board.utils import FenPlusHistory

from .game import Game
from .game_playing_status import GamePlayingStatus


@pytest.mark.parametrize(("use_rust_boards"), (True, False))
def test_game_rewind(use_rust_boards: bool) -> None:
    board: IBoard = create_board(
        use_rust_boards=use_rust_boards,
        fen_with_history=FenPlusHistory(current_fen=chess.STARTING_FEN),
    )

    game_playing_status: GamePlayingStatus = GamePlayingStatus()

    game: Game = Game(board=board, playing_status=game_playing_status)

    game.playing_status.pause()
    game.rewind_one_move()

    game.playing_status.play()
    game.play_move(move=game.board.get_move_key_from_uci("e2e4"))

    game.playing_status.pause()
    game.rewind_one_move()

    game.playing_status.play()
    game.play_move(move=game.board.get_move_key_from_uci("e2e4"))

    game.playing_status.play()
    game.play_move(move=game.board.get_move_key_from_uci("g8f6"))

    game.playing_status.pause()
    game.rewind_one_move()

    game.playing_status.play()
    game.play_move(move=game.board.get_move_key_from_uci("g8f6"))


if __name__ == "__main__":
    test_game_rewind(use_rust_boards=False)
    test_game_rewind(use_rust_boards=True)



"""
This module defines recommender rules for selecting moves in a tree-based move selector.

The recommender rules are implemented as data classes that define a `__call__` method. The `__call__` method takes a
`MoveAndValueTree` object and a random generator, and returns a recommended chess move.

The available recommender rule types are defined in the `RecommenderRuleTypes` enum.

The module also defines a `RecommenderRule` protocol that all recommender rule classes must implement.

Example usage:
    rule = AlmostEqualLogistic(type=RecommenderRuleTypes.AlmostEqualLogistic, temperature=0.5)
    move = rule(tree, random_generator)
"""

import random
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Literal, Protocol

import chipiron.players.move_selector.treevalue.trees as trees
from chipiron.environments.chess_env.move.imove import moveKey
from chipiron.players.boardevaluators.basic_evaluation import value_base
from chipiron.players.move_selector.treevalue.nodes.algorithm_node.algorithm_node import (
    AlgorithmNode,
)
from chipiron.players.move_selector.treevalue.nodes.utils import is_winning
from chipiron.utils.logger import chipiron_logger
from chipiron.utils.small_tools import softmax

if TYPE_CHECKING:
    from chipiron.players.move_selector.treevalue.nodes.itree_node import ITreeNode


class RecommenderRuleTypes(str, Enum):
    """
    Enum class that defines the available recommender rule types.
    """

    AlmostEqualLogistic = "almost_equal_logistic"
    Softmax = "softmax"


# theses are functions but i still use dataclasses instead
# of partial to be able to easily construct from yaml files using dacite


@dataclass
class AlmostEqualLogistic:
    """
    Recommender rule that selects the best move allowing for random choice for almost equally valued moves.
    """

    type: Literal[RecommenderRuleTypes.AlmostEqualLogistic]
    temperature: float

    def __call__(
        self, tree: trees.MoveAndValueTree, random_generator: random.Random
    ) -> moveKey:
        """
        Selects the best move from the tree, allowing for random choice for almost equally valued moves.

        Args:
            tree (trees.MoveAndValueTree): The move and value tree.
            random_generator (random.Random): The random generator.

        Returns:
            chess.Move: The recommended chess move.
        """
        # TODO this should be given at construction but postponed for now because of dataclasses
        # find the best first move allowing for random choice for almost equally valued moves.
        best_root_moves: list[moveKey] = (
            tree.root_node.minmax_evaluation.get_all_of_the_best_moves(
                how_equal="almost_equal_logistic"
            )
        )
        chipiron_logger.info(f"We have as bests: {[best for best in best_root_moves]}")
        best_move = random_generator.choice(best_root_moves)

        # assert isinstance(best_move, IMove)
        return best_move


@dataclass
class SoftmaxRule:
    """
    Recommender rule that selects the best move using the softmax function.
    """

    type: Literal[RecommenderRuleTypes.Softmax]
    temperature: float

    def __call__(
        self, tree: trees.MoveAndValueTree, random_generator: random.Random
    ) -> moveKey:
        """
        Selects the best move from the tree using the softmax function.

        Args:
            tree (trees.MoveAndValueTree): The move and value tree.
            random_generator (random.Random): The random generator.

        Returns:
            chess.Move: The recommended chess move.
        """
        # todo maybe there is a way to code this withtout the assert using the childrensorted value? or smth else
        values: list[float] = []
        for node in tree.root_node.moves_children.values():
            assert isinstance(node, AlgorithmNode)
            value = tree.root_node.minmax_evaluation.subjective_value_of(
                node.minmax_evaluation
            )
            values.append(float(value))

        softmax_ = list(softmax(values, self.temperature))

        move_as_list = random_generator.choices(
            list(tree.root_node.moves_children.keys()), weights=softmax_, k=1
        )
        best_move: moveKey = move_as_list[0]
        return best_move


AllRecommendFunctionsArgs = AlmostEqualLogistic | SoftmaxRule


@dataclass
class RecommenderRule(Protocol):
    """
    Protocol that all recommender rule classes must implement.
    """

    type: RecommenderRuleTypes

    def __call__(
        self, tree: trees.MoveAndValueTree, random_generator: random.Random
    ) -> moveKey:
        """
        Selects the best move from the tree.

        Args:
            tree (trees.MoveAndValueTree): The move and value tree.
            random_generator (random.Random): The random generator.

        Returns:
            chess.Move: The recommended chess move.
        """
        ...


def recommend_move_after_exploration_generic(
    recommend_move_after_exploration: AllRecommendFunctionsArgs,
    tree: trees.MoveAndValueTree,
    random_generator: random.Random,
) -> moveKey:
    """
    Recommends a move after exploration based on a generic rule.

    Args:
        recommend_move_after_exploration (AllRecommendFunctionsArgs): The recommend_move_after_exploration function.
        tree (trees.MoveAndValueTree): The move and value tree.
        random_generator (random.Random): The random number generator.

    Returns:
        chess.Move: The recommended move.

    """

    # if the situation is winning, we ask to play the move that is the most likely
    # to end the game fast by capturing pieces if possible
    is_winning_situation: bool = is_winning(
        node_minmax_evaluation=tree.root_node.minmax_evaluation,
        color=tree.root_node.board.turn,
    )
    over: bool = tree.root_node.is_over()
    if is_winning_situation and not over:
        # value of pieces of the opponent before the move
        value_father: int = value_base(
            board=tree.root_node.board, color=not tree.root_node.board.turn
        )

        child: ITreeNode | None
        best_value: int | None = None
        best_move: moveKey | None = None
        for move, child in tree.root_node.moves_children.items():
            assert isinstance(child, AlgorithmNode)

            # value of pieces of the opponent after that move
            value_child: int = value_base(board=child.board, color=child.board.turn)
            value: int = value_father - value_child

            still_wining_after_move: bool = is_winning(
                node_minmax_evaluation=child.minmax_evaluation,
                color=tree.root_node.board.turn,
            )
            if still_wining_after_move and (best_value is None or best_value < value):
                best_value = value
                best_move = move
        assert best_value is not None
        if best_value > 0:
            assert best_move is not None
            return best_move

    # base case
    return recommend_move_after_exploration(
        tree=tree, random_generator=random_generator
    )



from chipiron.players.boardevaluators.board_evaluation.board_evaluation import (
    PointOfView,
)
from chipiron.players.boardevaluators.neural_networks.output_converters.model_output_type import (
    ModelOutputType,
)
from chipiron.players.boardevaluators.neural_networks.output_converters.output_value_converter import (
    IdentityConverter,
    OutputValueConverter,
    PlayerToMoveValueToValueWhiteConverter,
)


def create_output_converter(model_output_type: ModelOutputType) -> OutputValueConverter:
    output_value_converter: OutputValueConverter
    point_of_view: PointOfView = model_output_type.point_of_view
    match point_of_view:
        case PointOfView.WHITE:
            output_value_converter = IdentityConverter()
        case PointOfView.PLAYER_TO_MOVE:
            output_value_converter = PlayerToMoveValueToValueWhiteConverter()
        case other:
            raise Exception(f"Not a valid output converter: {other} in file{__name__}")

    return output_value_converter



"""
Module for managing games.
"""

from .game.game_playing_status import GamePlayingStatus
from .game.observable_game_playing_status import ObservableGamePlayingStatus
from .match.match_manager import MatchManager

__all__ = ["MatchManager", "GamePlayingStatus", "ObservableGamePlayingStatus"]



from collections import Counter
from dataclasses import dataclass, field
from typing import Any, Iterator, Self, no_type_check

import chess
import shakmaty_python_binding

from chipiron.environments.chess_env.board.board_modification import (
    BoardModificationP,
    BoardModificationRust,
)
from chipiron.environments.chess_env.move import moveUci
from chipiron.environments.chess_env.move.imove import moveKey

from .iboard import (
    IBoard,
    LegalMoveKeyGeneratorP,
    boardKey,
    boardKeyWithoutCounters,
    compute_key,
)
from .utils import FenPlusHistory, fen


class LegalMoveKeyGeneratorRust(LegalMoveKeyGeneratorP):
    """LegalMoveKeyGeneratorRust is a Rust-compatible implementation of the LegalMoveKeyGeneratorP interface."""

    # whether to sort the legal_moves by their respective uci for easy comparison of various implementations
    sort_legal_moves: bool

    all_generated_keys: list[moveKey] | None

    chess_rust_binding: shakmaty_python_binding.MyChess

    def get_uci_from_move_key(self, move_key: moveKey) -> moveUci:
        """Returns the UCI string corresponding to the given move key.
        Args:
            move_key (moveKey): The move key to convert to UCI.
        Returns:
            moveUci: The UCI string corresponding to the given move key."""
        assert self.generated_moves is not None
        chess_move: shakmaty_python_binding.MyMove = self.generated_moves[move_key]
        return chess_move.uci()

    def __init__(
        self,
        sort_legal_moves: bool,
        chess_rust_binding: shakmaty_python_binding.MyChess,
        generated_moves: list[shakmaty_python_binding.MyMove] | None = None,
    ):
        """LegalMoveKeyGeneratorRust is a Rust-compatible implementation of the LegalMoveKeyGeneratorP interface.

        Args:
            sort_legal_moves (bool): Whether to sort the legal moves.
            chess_rust_binding (shakmaty_python_binding.MyChess): The Rust chess binding.
            generated_moves (list[shakmaty_python_binding.MyMove] | None, optional): The generated moves. Defaults to None.
        """
        self.chess_rust_binding = chess_rust_binding
        self.generated_moves = generated_moves
        if generated_moves is not None:
            self.number_moves = len(generated_moves)
            self.it: Iterator[int] = iter(range(self.number_moves))
            self.all_generated_keys = list(range(self.number_moves))
            if sort_legal_moves:

                def f(i: int) -> moveUci:
                    assert self.generated_moves is not None
                    return self.generated_moves[i].uci()

                self.all_generated_keys = sorted(
                    list(range(self.number_moves)),
                    # key=lambda i: self.generated_moves[i].uci()
                    key=f,
                )
            else:
                self.all_generated_keys = list(range(self.number_moves))
        else:
            self.all_generated_keys = None
        self.sort_legal_moves = sort_legal_moves

    @property
    def fen(self) -> fen:
        return self.chess_rust_binding.fen()

    def reset(self, generated_moves: list[shakmaty_python_binding.MyMove]) -> None:
        self.generated_moves = generated_moves
        self.number_moves = len(generated_moves)
        self.it = iter(range(self.number_moves))
        self.all_generated_keys = list(range(self.number_moves))

    def copy_with_reset(self) -> "LegalMoveKeyGeneratorRust":
        legal_move_copy = LegalMoveKeyGeneratorRust(
            chess_rust_binding=self.chess_rust_binding,
            generated_moves=None,
            sort_legal_moves=self.sort_legal_moves,
        )
        return legal_move_copy

    def set_legal_moves(
        self, generated_moves: list[shakmaty_python_binding.MyMove]
    ) -> None:
        self.generated_moves = generated_moves
        self.number_moves = len(generated_moves)

    def __iter__(self) -> Iterator[moveKey]:
        if self.generated_moves is None:
            self.generated_moves = self.chess_rust_binding.legal_moves()
        if self.sort_legal_moves:
            assert self.generated_moves is not None

            def f(i: int) -> moveUci:
                assert self.generated_moves is not None
                return self.generated_moves[i].uci()

            self.it = iter(
                sorted(
                    list(range(self.number_moves)),
                    # key=lambda i: self.generated_moves[i].uci()
                    key=f,
                )
            )
        else:
            self.it = iter(range(self.number_moves))
        return self

    def __next__(self) -> moveKey:
        return self.it.__next__()

    def copy(
        self, copied_chess_rust_binding: shakmaty_python_binding.MyChess | None = None
    ) -> "LegalMoveKeyGeneratorRust":
        if copied_chess_rust_binding is None:
            copied_chess_rust_binding_ = self.chess_rust_binding
        else:
            copied_chess_rust_binding_ = copied_chess_rust_binding
        legal_move_copy = LegalMoveKeyGeneratorRust(
            chess_rust_binding=copied_chess_rust_binding_,
            generated_moves=(
                self.generated_moves.copy()
                if self.generated_moves is not None
                else None
            ),
            sort_legal_moves=self.sort_legal_moves,
        )
        if self.all_generated_keys is not None:
            legal_move_copy.all_generated_keys = self.all_generated_keys.copy()
        else:
            legal_move_copy.all_generated_keys = legal_move_copy.all_generated_keys
        return legal_move_copy

    def get_all(self) -> list[moveKey]:
        if self.generated_moves is None:
            self.generated_moves = self.chess_rust_binding.legal_moves()
            self.number_moves = len(self.generated_moves)
            self.all_generated_keys = None

        if self.all_generated_keys is None:
            if self.sort_legal_moves:

                def f(i: int) -> moveUci:
                    assert self.generated_moves is not None
                    return self.generated_moves[i].uci()

                s = sorted(
                    list(range(self.number_moves)),
                    # key=lambda i: self.generated_moves[i].uci()
                    key=f,
                )
                return s
            else:
                return list(range(self.number_moves))
        else:
            return self.all_generated_keys

    def more_than_one_move(self) -> bool:
        if self.generated_moves is None:
            self.generated_moves = self.chess_rust_binding.legal_moves()
            self.number_moves = len(self.generated_moves)
            self.all_generated_keys = None
        return len(self.generated_moves) > 0


# todo implement rewind (and a test for it)


@dataclass
class RustyBoardChi(IBoard):
    """
    Rusty Board Chipiron
    object that describes the current board. it wraps the chess Board from the chess package so it can have more in it
    but im not sure its really necessary.i keep it for potential usefulness

    This is the Rust version for speedy execution
    It is based on the binding library shakmaty_python_binding to use the rust library shakmaty
    """

    # the shakmaty implementation of the board that we wrap here
    chess_: shakmaty_python_binding.MyChess

    compute_board_modification: bool

    # to count the number of occurrence of each board to be able to compute
    # three-fold repetition as shakmaty does not do it atm
    rep_to_count: Counter[boardKeyWithoutCounters]

    fast_representation_: boardKey

    # storing the info here for fast access as it seems calls to rust bingings can be costy
    pawns_: int
    kings_: int
    queens_: int
    rooks_: int
    bishops_: int
    knights_: int
    white_: int
    black_: int
    turn_: bool
    ep_square_: int | None
    promoted_: int
    castling_rights_: int

    legal_moves_: LegalMoveKeyGeneratorRust

    # the move history is kept here because shakmaty_python_binding.MyChess does not have a move stack at the moment
    move_stack: list[moveUci] = field(default_factory=lambda: list[moveUci]())

    # the board history is kept here because shakmaty_python_binding.MyChess does not have a board stack at the moment
    # board_stack: list[chess._BoardState] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.rep_to_count[self.fast_representation_without_counters] = 1

    def __str__(self) -> str:
        """
        Returns a string representation of the board.

        Returns:
            str: A string representation of the board.
        """
        return self.fen

    def play_min_2(self, move: shakmaty_python_binding.MyMove) -> None:
        # _str, ply, turn, is_game_over = self.chess_.play_and_return(move)
        (
            self.castling_rights_,
            self.pawns_,
            self.knights_,
            self.bishops_,
            self.rooks_,
            self.queens_,
            self.kings_,
            self.white_,
            self.black_,
            turn_int,
            ep_square_int,
            self.promoted_,
        ) = self.chess_.play_and_return_o(move)
        self.turn_ = bool(turn_int)
        if ep_square_int == -1:
            self.ep_square_ = None
        else:
            self.ep_square_ = ep_square_int

    def play_min_3(self, move: shakmaty_python_binding.MyMove) -> BoardModificationRust:
        (
            (
                self.castling_rights_,
                self.pawns_,
                self.knights_,
                self.bishops_,
                self.rooks_,
                self.queens_,
                self.kings_,
                self.white_,
                self.black_,
                turn_int,
                ep_square_int,
                self.promoted_,
            ),
            appearances,
            removals,
        ) = self.chess_.play_and_return_modifications(move)
        self.turn_ = bool(turn_int)
        if ep_square_int == -1:
            self.ep_square_ = None
        else:
            self.ep_square_ = ep_square_int

        board_modifications: BoardModificationRust = self.convert(appearances, removals)
        return board_modifications

    def convert(
        self,
        appearances: set[tuple[int, int, int]],
        removals: set[tuple[int, int, int]],
    ) -> BoardModificationRust:
        board_modifications: BoardModificationRust = BoardModificationRust(
            appearances_=appearances, removals_=removals
        )

        return board_modifications

    def play_move(
        self, move: shakmaty_python_binding.MyMove
    ) -> BoardModificationP | None:
        """
        Plays a move on the board and returns the board modification.

        Args:
            move: The move to play.

        Returns:
            The board modification resulting from the move or None.
        """
        # todo: illegal moves seem accepted, do we care? if we dont write it in the doc
        # assert self.board.is_legal(move)
        #
        board_modifications: BoardModificationRust | None = None

        if self.compute_board_modification:
            if True:
                board_modifications = self.play_min_3(move)

            # else:
            #     previous_pawns = self.pawns_
            #     previous_kings = self.kings_
            #     previous_queens = self.queens_
            #     previous_rooks = self.rooks_
            #     previous_bishops = self.bishops_
            #     previous_knights = self.knights_
            #     previous_occupied_white = self.white_
            #     previous_occupied_black = self.black_
            #
            #     self.play_min_2(move)
            #
            #     new_pawns = self.pawns_
            #     new_kings = self.kings_
            #     new_queens = self.queens_
            #     new_rooks = self.rooks_
            #     new_bishops = self.bishops_
            #     new_knights = self.knights_
            #     new_occupied_white = self.white_
            #     new_occupied_black = self.black_
            #
            #     board_modifications = compute_modifications(
            #         previous_bishops=previous_bishops,
            #         previous_pawns=previous_pawns,
            #         previous_kings=previous_kings,
            #         previous_knights=previous_knights,
            #         previous_queens=previous_queens,
            #         previous_occupied_white=previous_occupied_white,
            #         previous_rooks=previous_rooks,
            #         previous_occupied_black=previous_occupied_black,
            #         new_kings=new_kings,
            #         new_bishops=new_bishops,
            #         new_pawns=new_pawns,
            #         new_queens=new_queens,
            #         new_rooks=new_rooks,
            #         new_knights=new_knights,
            #         new_occupied_black=new_occupied_black,
            #         new_occupied_white=new_occupied_white
            #     )
        else:
            self.play_min_2(move)

        # update after move
        self.legal_moves_ = (
            self.legal_moves_.copy_with_reset()
        )  # the legals moves needs to be recomputed as the board has changed

        fast_representation: boardKey = compute_key(
            pawns=self.pawns_,
            knights=self.knights_,
            bishops=self.bishops_,
            rooks=self.rooks_,
            queens=self.queens_,
            kings=self.kings_,
            turn=self.turn_,
            castling_rights=self.castling_rights_,
            ep_square=self.ep_square_,
            white=self.white_,
            black=self.black_,
            promoted=self.promoted_,
            fullmove_number=self.chess_.fullmove_number(),
            halfmove_clock=self.chess_.halfmove_clock(),
        )
        self.fast_representation_ = fast_representation
        self.rep_to_count.update([self.fast_representation_without_counters])
        self.move_stack.append(move.uci())
        # self.board_stack.append(chess._BoardState())

        return board_modifications

    def play_move_uci(self, move_uci: moveUci) -> BoardModificationP | None:
        chess_move: shakmaty_python_binding.MyMove = shakmaty_python_binding.MyMove(
            uci=move_uci, my_chess=self.chess_
        )
        return self.play_move(move=chess_move)

    # todo look like this function might move to iboard when the dust settle
    def play_move_key(self, move: moveKey) -> BoardModificationP | None:
        assert self.legal_moves_.generated_moves is not None
        my_move: shakmaty_python_binding.MyMove = self.legal_moves_.generated_moves[
            move
        ]
        return self.play_move(move=my_move)

    def ply(self) -> int:
        """
        Returns the number of half-moves (plies) that have been played on the board.

        :return: The number of half-moves played on the board.
        :rtype: int
        """
        ply: int = self.chess_.ply()
        return ply

    @property
    def turn(self) -> chess.Color:
        """
        Get the current turn color.

        Returns:
            chess.Color: The color of the current turn.
        """
        # return bool(self.chess_.turn())
        return self.turn_

    @no_type_check
    def is_game_over(self) -> bool:
        """
        Check if the game is over.

        Returns:
            bool: True if the game is over, False otherwise.
        """
        claim_draw: bool = True if len(self.move_stack) >= 5 else False
        three_fold_repetition: bool = (
            max(self.rep_to_count.values()) > 2 if claim_draw else False
        )
        # todo check the move stack : check for repetition as the rust version not do it
        # todo remove this hasatrribute at some point

        chess_game_over: bool = False

        if hasattr(self, "is_game_over_"):
            attr_value: Any = self.is_game_over_
            chess_game_over = bool(attr_value)
        else:
            engine_result: Any = self.chess_.is_game_over()
            chess_game_over = bool(engine_result)

        return three_fold_repetition or chess_game_over

    def copy(self, stack: bool, deep_copy_legal_moves: bool = True) -> Self:
        """
        Create a copy of the current board.

        Args:
            stack (bool): Whether to copy the move stack as well.

        Returns:
            RustyBoardChi: A new instance of the BoardChi class with the copied board.
        """
        chess_copy: shakmaty_python_binding.MyChess = self.chess_.copy()
        move_stack_ = self.move_stack.copy() if stack else []
        # board_stack_ = self.board_stack.copy() if stack else []

        legal_moves_copy: LegalMoveKeyGeneratorRust
        if deep_copy_legal_moves:
            legal_moves_copy = self.legal_moves_.copy(
                copied_chess_rust_binding=chess_copy
            )
        else:
            legal_moves_copy = self.legal_moves_
            legal_moves_copy.chess_rust_binding = chess_copy

        return type(self)(
            chess_=chess_copy,
            move_stack=move_stack_,
            # board_stack=board_stack_
            compute_board_modification=self.compute_board_modification,
            rep_to_count=self.rep_to_count.copy(),
            fast_representation_=self.fast_representation_,
            pawns_=self.pawns_,
            knights_=self.knights_,
            kings_=self.kings_,
            rooks_=self.rooks_,
            queens_=self.queens_,
            bishops_=self.bishops_,
            black_=self.black_,
            white_=self.white_,
            turn_=self.turn_,
            ep_square_=self.ep_square_,
            promoted_=self.promoted_,
            castling_rights_=self.castling_rights_,
            legal_moves_=legal_moves_copy,
        )

    @property
    def legal_moves(self) -> LegalMoveKeyGeneratorRust:
        # todo minimize this call and understand when the role of the variable all legal move generated
        return self.legal_moves_

    def number_of_pieces_on_the_board(self) -> int:
        """
        Returns the number of pieces currently on the board.

        Returns:
            int: The number of pieces on the board.
        """
        return self.chess_.number_of_pieces_on_the_board()

    @property
    def fen(self) -> fen:
        """
        Returns the Forsyth-Edwards Notation (FEN) representation of the chess board.

        :return: The FEN string representing the current state of the board.
        """
        return self.chess_.fen()

    def piece_at(self, square: chess.Square) -> chess.Piece | None:
        """
        Returns the piece at the specified square on the chess board.

        Args:
            square (chess.Square): The square on the chess board.

        Returns:
            chess.Piece | None: The piece at the specified square, or None if there is no piece.

        """
        piece_or_none = self.chess_.piece_at(square)
        piece: chess.Piece | None
        if piece_or_none is None:
            piece = None
        else:
            piece = chess.Piece(piece_type=piece_or_none[1], color=piece_or_none[0])
        return piece

    def piece_map(self) -> dict[chess.Square, tuple[int, bool]]:
        dict_raw = self.chess_.piece_map()
        return dict_raw

    def has_kingside_castling_rights(self, color: chess.Color) -> bool:
        """
        Check if the specified color has kingside castling rights.

        Args:
            color (chess.Color): The color to check for kingside castling rights.

        Returns:
            bool: True if the specified color has kingside castling rights, False otherwise.
        """
        return self.chess_.has_kingside_castling_rights(color)

    def has_queenside_castling_rights(self, color: chess.Color) -> bool:
        """
        Check if the specified color has queenside castling rights.

        Args:
            color (chess.Color): The color to check for queenside castling rights.

        Returns:
            bool: True if the specified color has kingside castling rights, False otherwise.
        """
        return self.chess_.has_queenside_castling_rights(color)

    def print_chess_board(self) -> str:
        """
        Prints the current state of the chess board.

        This method prints the current state of the chess board, including the position of all the pieces.
        It also prints the FEN (Forsyth–Edwards Notation) representation of the board.

        Returns:
            None
        """
        return str(self.chess_.fen())

    def tell_result(self) -> None: ...

    @property
    def move_history_stack(self) -> list[moveUci]:
        return self.move_stack

    # @property
    # def board_history_stack(self) -> list[chess._BoardState]:
    #    return self.board_stack

    def dump(self, file: Any) -> None:
        """Dumps the current state of the board to the specified file."""
        ...

    def is_attacked(self, a_color: chess.Color) -> bool:
        """Check if any piece of the color `a_color` is attacked.

        Args:
            a_color (chess.Color): The color of the pieces to check.

        Returns:
            bool: True if any piece of the specified color is attacked, False otherwise.
        """
        return self.chess_.is_attacked(a_color)

    @property
    def pawns(self) -> chess.Bitboard:
        return self.chess_.pawns()

    @property
    def knights(self) -> chess.Bitboard:
        return self.knights_
        # return self.chess_.knights()

    @property
    def bishops(self) -> chess.Bitboard:
        return self.bishops_
        #        return self.chess_.bishops()

    @property
    def rooks(self) -> chess.Bitboard:
        return self.rooks_
        # return self.chess_.rooks()

    @property
    def queens(self) -> chess.Bitboard:
        return self.queens_

    #        return self.chess_.queens()

    @property
    def kings(self) -> chess.Bitboard:
        return self.kings_

    #        return self.chess_.kings()

    @property
    def white(self) -> chess.Bitboard:
        return self.white_
        # return self.chess_.white()

    @property
    def black(self) -> chess.Bitboard:
        return self.black_
        # return self.chess_.black()

    @property
    def occupied(self) -> chess.Bitboard:
        return self.chess_.occupied()

    def result(self, claim_draw: bool = False) -> str:
        claim_draw_: bool = True if len(self.move_stack) >= 5 and claim_draw else False
        three_fold_repetition: bool = (
            max(self.rep_to_count.values()) > 2 if claim_draw_ else False
        )

        if three_fold_repetition:
            return "1/2-1/2"
        else:
            return self.chess_.result()

    @property
    def castling_rights(self) -> chess.Bitboard:
        # return self.chess_.castling_rights()
        return self.castling_rights_

    def termination(self) -> None:
        return None

    def occupied_color(self, color: chess.Color) -> chess.Bitboard:
        if color == chess.WHITE:
            return self.chess_.white()
        else:
            return self.chess_.black()

    @property
    def halfmove_clock(self) -> int:
        return self.chess_.halfmove_clock()

    @property
    def promoted(self) -> chess.Bitboard:
        return self.promoted_

    @property
    def fullmove_number(self) -> int:
        return self.chess_.fullmove_number()

    @property
    def ep_square(self) -> int | None:
        return self.ep_square_

    def is_zeroing(self, move: moveKey) -> bool:
        """Check if a move is a zeroing move (i.e., checks if the given move is a capture or pawn move.

        Args:
            move (moveKey): The move to check.

        Returns:
            bool: True if the move is a zeroing move, False otherwise.
        """
        assert self.legal_moves_.generated_moves is not None
        chess_move: shakmaty_python_binding.MyMove = self.legal_moves_.generated_moves[
            move
        ]
        return chess_move.is_zeroing()

    def into_fen_plus_history(self) -> FenPlusHistory:
        return FenPlusHistory(
            current_fen=self.fen, historical_moves=self.move_history_stack
        )



"""
This module defines stopping criteria for a move selector in a game tree.

The stopping criteria determine when the move selector should stop exploring the game tree and make a decision.

The module includes the following classes:

- StoppingCriterion: The general stopping criterion class.
- TreeMoveLimit: A stopping criterion based on a tree move limit.
- DepthLimit: A stopping criterion based on a depth limit.

It also includes helper classes and functions for creating and managing stopping criteria.
"""

import abc
from dataclasses import dataclass
from enum import Enum
from typing import Callable, Literal, Protocol, runtime_checkable

from chipiron.players.move_selector.treevalue import node_selector as node_sel
from chipiron.players.move_selector.treevalue.trees import MoveAndValueTree


@runtime_checkable
class DepthToExpendP(Protocol):
    """
    Protocol for objects that provide the current depth to expand.

    This protocol defines a single method `get_current_depth_to_expand` that should be implemented by classes
    that want to provide the current depth to expand.

    Attributes:
        None

    Methods:
        get_current_depth_to_expand: Returns the current depth to expand as an integer.

    Examples:
        >>> class MyDepthToExpend(DepthToExpendP):
        ...     def get_current_depth_to_expand(self) -> int:
        ...         return 5
        ...
        >>> obj = MyDepthToExpend()
        >>> obj.get_current_depth_to_expand()
        5
    """

    def get_current_depth_to_expand(self) -> int:
        """
        Returns the current depth to expand as an integer.

        Returns:
            The current depth to expand.

        Raises:
            None
        """
        ...


class StoppingCriterionTypes(str, Enum):
    """
    Enum class representing different types of stopping criteria for tree value calculation.
    """

    DEPTH_LIMIT = "depth_limit"
    TREE_MOVE_LIMIT = "tree_move_limit"


@dataclass
class StoppingCriterionArgs:
    """
    Represents the arguments for a stopping criterion.

    Attributes:
        type (StoppingCriterionTypes): The type of stopping criterion.
    """

    type: StoppingCriterionTypes


class ProgressMonitorP(Protocol):
    """
    The general stopping criterion Protocol
    """

    def should_we_continue(self, tree: MoveAndValueTree) -> bool:
        """
        Asking should we continue

        Returns:
            boolean of should we continue
        """
        ...

    def respectful_opening_instructions(
        self, opening_instructions: node_sel.OpeningInstructions, tree: MoveAndValueTree
    ) -> node_sel.OpeningInstructions:
        """
        Ensures the opening request do not exceed the stopping criterion


        """
        ...

    def get_string_of_progress(self, tree: MoveAndValueTree) -> str:
        """
        Returns a string representation of the progress made by the stopping criterion.

        Args:
            tree (MoveAndValueTree): The move and value tree.

        Returns:
            str: A string representation of the progress.
        """
        ...

    def get_percent_of_progress(
        self, tree: MoveAndValueTree, notify_function: Callable[[int], None] | None
    ) -> str: ...


class ProgressMonitor:
    """
    The general stopping criterion base class
    """

    def should_we_continue(self, tree: MoveAndValueTree) -> bool:
        """
        Asking should we continue

        Returns:
            boolean of should we continue
        """
        if tree.is_over():
            return False
        return True

    def respectful_opening_instructions(
        self, opening_instructions: node_sel.OpeningInstructions, tree: MoveAndValueTree
    ) -> node_sel.OpeningInstructions:
        """
        Ensures the opening request do not exceed the stopping criterion


        """
        return opening_instructions

    def get_string_of_progress(self, tree: MoveAndValueTree) -> str:
        """
        Returns a string representation of the progress made by the stopping criterion.

        Args:
            tree (MoveAndValueTree): The move and value tree.

        Returns:
            str: A string representation of the progress.
        """
        return ""

    @abc.abstractmethod
    def get_percent_of_progress(self, tree: MoveAndValueTree) -> int: ...

    def notify_percent_progress(
        self,
        tree: MoveAndValueTree,
        notify_percent_function: Callable[[int], None] | None,
    ) -> None:
        percent_progress: int = self.get_percent_of_progress(tree=tree)

        if notify_percent_function is not None:
            notify_percent_function(percent_progress)


@dataclass
class TreeMoveLimitArgs:
    """Arguments for the tree move limit stopping criterion."""

    type: Literal[StoppingCriterionTypes.TREE_MOVE_LIMIT]
    tree_move_limit: int


class TreeMoveLimit(ProgressMonitor):
    """
    The stopping criterion based on a tree move limit
    """

    tree_move_limit: int

    def __init__(self, tree_move_limit: int) -> None:
        self.tree_move_limit = tree_move_limit

    def should_we_continue(self, tree: MoveAndValueTree) -> bool:
        continue_base: bool = super().should_we_continue(tree=tree)

        should_we: bool
        if not continue_base:
            should_we = continue_base
        else:
            should_we = tree.move_count < self.tree_move_limit
        return should_we

    def respectful_opening_instructions(
        self, opening_instructions: node_sel.OpeningInstructions, tree: MoveAndValueTree
    ) -> node_sel.OpeningInstructions:
        """
        Ensures the opening request do not exceed the stopping criterion


        """
        opening_instructions_subset: node_sel.OpeningInstructions = (
            node_sel.OpeningInstructions()
        )
        opening_instructions.pop_items(
            popped=opening_instructions_subset,
            how_many=self.tree_move_limit - tree.move_count,
        )
        return opening_instructions_subset

    def get_string_of_progress(self, tree: MoveAndValueTree) -> str:
        """
        compute the string that display the progress in the terminal

        Returns:
            a string that display the progress in the terminal
        """
        return (
            f"========= tree move counting: {tree.move_count} out of {self.tree_move_limit}"
            f" |  {tree.move_count / self.tree_move_limit:.0%}"
        )

    def get_percent_of_progress(
        self,
        tree: MoveAndValueTree,
    ) -> int:
        percent: int = int(tree.move_count / self.tree_move_limit * 100)
        return percent


@dataclass
class DepthLimitArgs:
    """
    Arguments for the depth limit stopping criterion.

    Attributes:
        depth_limit (int): The maximum depth allowed for the search.
    """

    type: Literal[StoppingCriterionTypes.DEPTH_LIMIT]
    depth_limit: int


class DepthLimit(ProgressMonitor):
    """
    The stopping criterion based on a depth limit
    """

    depth_limit: int
    node_selector: DepthToExpendP

    def __init__(self, depth_limit: int, node_selector: DepthToExpendP) -> None:
        """
        Initializes a StoppingCriterion object.

        Args:
            depth_limit (int): The maximum depth to search in the tree.
            node_selector (DepthToExpendP): The node selector used to determine which nodes to expand.

        Returns:
            None
        """
        self.depth_limit = depth_limit
        self.node_selector = node_selector

    def should_we_continue(self, tree: MoveAndValueTree) -> bool:
        """
        Determines whether the search should continue expanding nodes in the tree.

        Args:
            tree (MoveAndValueTree): The tree containing the moves and their corresponding values.

        Returns:
            bool: True if the search should continue, False otherwise.
        """
        continue_base = super().should_we_continue(tree=tree)
        if not continue_base:
            return continue_base
        return self.node_selector.get_current_depth_to_expand() < self.depth_limit

    def get_string_of_progress(self, tree: MoveAndValueTree) -> str:
        """
        compute the string that display the progress in the terminal

        Returns:
            a string that display the progress in the terminal
        """
        return (
            "========= tree move counting: "
            + str(tree.move_count)
            + " | Depth: "
            + str(self.node_selector.get_current_depth_to_expand())
            + " out of "
            + str(self.depth_limit)
        )

    def get_percent_of_progress(
        self,
        tree: MoveAndValueTree,
    ) -> int:
        # todo this percent is not precise
        percent: int = int(
            self.node_selector.get_current_depth_to_expand() / self.depth_limit * 100
        )
        return percent


AllStoppingCriterionArgs = TreeMoveLimitArgs | DepthLimitArgs


def create_stopping_criterion(
    args: AllStoppingCriterionArgs, node_selector: node_sel.NodeSelector
) -> ProgressMonitor:
    """
    creating the stopping criterion

    Args:
        args:
        node_selector:

    Returns:
        A stopping criterion

    """
    stopping_criterion: ProgressMonitor

    match args.type:
        case StoppingCriterionTypes.DEPTH_LIMIT:
            assert isinstance(node_selector, DepthToExpendP)
            assert isinstance(args, DepthLimitArgs)
            stopping_criterion = DepthLimit(
                depth_limit=args.depth_limit, node_selector=node_selector
            )
        case StoppingCriterionTypes.TREE_MOVE_LIMIT:
            assert isinstance(args, TreeMoveLimitArgs)

            stopping_criterion = TreeMoveLimit(tree_move_limit=args.tree_move_limit)
        case _:
            raise ValueError(
                f"stopping criterion builder: can not find {args.type} in file {__name__}"
            )

    return stopping_criterion



"""
This module provides functions for visualizing and saving tree structures.

The functions in this module allow for the visualization of tree structures using the Graphviz library.
It provides a way to display the tree structure as a graph and save it as a PDF file.
Additionally, it provides a function to save the raw data of the tree structure to a file using pickle.

Functions:
- add_dot(dot: Digraph, treenode: ITreeNode) -> None: Adds nodes and edges to the graph representation of the tree.
- display_special(node: ITreeNode, format: str, index: dict[chess.Move, str]) -> Digraph: Displays a special
representation of the tree with additional information.
- display(tree: MoveAndValueTree, format_str: str) -> Digraph: Displays the tree structure as a graph.
- save_pdf_to_file(tree: MoveAndValueTree) -> None: Saves the tree structure as a PDF file.
- save_raw_data_to_file(tree: MoveAndValueTree, count: str = '#') -> None: Saves the raw data of the tree
structure to a file.
"""

import pickle

from graphviz import Digraph

from chipiron.environments.chess_env.move.imove import moveKey
from chipiron.players.move_selector.treevalue.nodes import ITreeNode
from chipiron.players.move_selector.treevalue.nodes.algorithm_node.algorithm_node import (
    AlgorithmNode,
)

from .move_and_value_tree import MoveAndValueTree


def add_dot(dot: Digraph, treenode: ITreeNode) -> None:
    """
    Adds a node and edges to the given Dot graph based on the provided tree node.

    Args:
        dot (Digraph): The Dot graph to add the node and edges to.
        treenode (ITreeNode): The tree node to visualize.

    Returns:
        None
    """
    nd = treenode.dot_description()
    dot.node(str(treenode.id), nd)
    move: moveKey
    for _, move in enumerate(treenode.moves_children):
        if treenode.moves_children[move] is not None:
            child = treenode.moves_children[move]
            if child is not None:
                cdd = str(child.id)
                dot.edge(
                    str(treenode.id),
                    cdd,
                    str(treenode.board.get_uci_from_move_key(move_key=move)),
                )
                add_dot(dot, child)


def display_special(
    node: ITreeNode, format_str: str, index: dict[moveKey, str]
) -> Digraph:
    """
    Display a special visualization of a tree node and its children.

    Args:
        node (ITreeNode): The tree node to display.
        format_str (str): The format of the output graph (e.g., 'png', 'pdf', 'svg').
        index (Dict[chess.Move, str]): A dictionary mapping chess moves to their descriptions.

    Returns:
        Digraph: The graph representing the tree visualization.

    Raises:
        AssertionError: If the child node is None or if the parent node is not an AlgorithmNode.
    """
    dot = Digraph(format=format_str)
    print(";;;", type(node))
    nd = node.dot_description()
    dot.node(str(node.id), nd)
    sorted_moves = [(str(move), move) for move in node.moves_children.keys()]
    sorted_moves.sort()
    for move_key in sorted_moves:
        move = move_key[1]
        if node.moves_children[move] is not None:
            child = node.moves_children[move]
            assert child is not None
            assert isinstance(node, AlgorithmNode)

            cdd = str(child.id)
            edge_description = (
                index[move]
                + "|"
                + str(node.board.get_uci_from_move_key(move_key=move))
                + "|"
                + node.minmax_evaluation.description_tree_visualizer_move(child)
            )
            dot.edge(str(node.id), cdd, edge_description)
            dot.node(str(child.id), child.dot_description())
            print("--move:", edge_description)
            print("--child:", child.dot_description())
    return dot


def display(tree: MoveAndValueTree, format_str: str) -> Digraph:
    """
    Display the move and value tree using graph visualization.

    Args:
        tree (MoveAndValueTree): The move and value tree to be displayed.
        format_str (str): The format of the output graph (e.g., 'png', 'pdf', 'svg').

    Returns:
        Digraph: The graph representation of the move and value tree.
    """
    dot = Digraph(format=format_str)
    add_dot(dot, tree.root_node)
    return dot


def save_pdf_to_file(tree: MoveAndValueTree) -> None:
    """
    Saves the visualization of a tree as a PDF file.

    Args:
        tree (MoveAndValueTree): The tree to be visualized and saved.

    Returns:
        None
    """
    dot = display(tree=tree, format_str="pdf")
    round_ = len(tree.root_node.board.move_history_stack) + 2
    color = "white" if tree.root_node.player_to_move else "black"
    dot.render(
        "chipiron/runs/treedisplays/TreeVisual_" + str(int(round_ / 2)) + color + ".pdf"
    )


def save_raw_data_to_file(tree: MoveAndValueTree, count: str = "#") -> None:
    """
    Save raw data of a MoveAndValueTree to a file.

    Args:
        tree (MoveAndValueTree): The MoveAndValueTree object to save.
        count (str, optional): A string to append to the filename. Defaults to '#'.

    Returns:
        None
    """
    round_ = len(tree.root_node.board.move_history_stack) + 2
    color = "white" if tree.root_node.player_to_move else "black"
    filename = (
        "chipiron/debugTreeData_"
        + str(int(round_ / 2))
        + color
        + "-"
        + str(count)
        + ".td"
    )

    import sys

    sys.setrecursionlimit(100000)
    with open(filename, "wb") as f:
        pickle.dump([tree.descendants, tree.root_node], f)



"""
This module contains the implementation of the TreeAndValueMoveSelector class, which is responsible for selecting moves
based on a tree and value strategy.

The TreeAndValueMoveSelector class uses a tree-based approach to explore possible moves and select the best move based on
a value function. It utilizes a tree manager, a tree factory, stopping criterion arguments, a node selector factory, a
random generator, and recommendation functions to guide the move selection process.

The TreeAndValueMoveSelector class provides the following methods:
- select_move: Selects the best move based on the tree and value strategy.
- print_info: Prints information about the move selector type.
"""

import queue
import random
from dataclasses import dataclass

import chipiron.environments.chess_env.board as boards
from chipiron.players.move_selector.move_selector import MoveRecommendation
from chipiron.players.move_selector.treevalue.progress_monitor.progress_monitor import (
    AllStoppingCriterionArgs,
)
from chipiron.players.move_selector.treevalue.search_factory import NodeSelectorFactory
from chipiron.utils import Seed
from chipiron.utils.dataclass import IsDataclass

from . import recommender_rule
from . import tree_manager as tree_man
from .tree_exploration import TreeExploration, create_tree_exploration
from .trees.factory import MoveAndValueTreeFactory


@dataclass
class TreeAndValueMoveSelector:
    """
    The TreeAndValueMoveSelector class is responsible for selecting moves based on a tree and value strategy.

    Attributes:
    - tree_manager: The tree manager responsible for managing the algorithm nodes.
    - tree_factory: The tree factory responsible for creating move and value trees.
    - stopping_criterion_args: The stopping criterion arguments used to determine when to stop the tree exploration.
    - node_selector_create: The node selector factory used to create node selectors for tree exploration.
    - random_generator: The random generator used for randomization during tree exploration.
    - recommend_move_after_exploration: The recommendation functions used to recommend a move after tree exploration.
    """

    # pretty empty class but might be useful when dealing with multi round and time , no?

    tree_manager: tree_man.AlgorithmNodeTreeManager
    tree_factory: MoveAndValueTreeFactory
    stopping_criterion_args: AllStoppingCriterionArgs
    node_selector_create: NodeSelectorFactory
    random_generator: random.Random
    recommend_move_after_exploration: recommender_rule.AllRecommendFunctionsArgs
    queue_progress_player: queue.Queue[IsDataclass] | None

    def select_move(self, board: boards.IBoard, move_seed: Seed) -> MoveRecommendation:
        """
        Selects the best move based on the tree and value strategy.

        Args:
        - board: The current board state.
        - move_seed: The seed used for randomization during move selection.

        Returns:
        - The recommended move based on the tree and value strategy.
        """
        tree_exploration: TreeExploration = self.create_tree_exploration(board=board)
        self.random_generator.seed(move_seed)

        move_recommendation: MoveRecommendation = tree_exploration.explore(
            random_generator=self.random_generator
        ).move_recommendation

        return move_recommendation

    def create_tree_exploration(
        self,
        board: boards.IBoard,
    ) -> TreeExploration:
        tree_exploration: TreeExploration = create_tree_exploration(
            tree_manager=self.tree_manager,
            node_selector_create=self.node_selector_create,
            starting_board=board,
            tree_factory=self.tree_factory,
            stopping_criterion_args=self.stopping_criterion_args,
            recommend_move_after_exploration=self.recommend_move_after_exploration,
            queue_progress_player=self.queue_progress_player,
        )
        return tree_exploration

    def print_info(self) -> None:
        """
        Prints information about the move selector type.
        """
        print("type: Tree and Value")



"""
Module for the SyzygyTable class.
"""

from typing import Protocol

import chess

from chipiron.environments.chess_env.board import IBoard
from chipiron.environments.chess_env.move.imove import moveKey
from chipiron.players.boardevaluators.over_event import HowOver, OverTags, Winner


class SyzygyTable[T_Board: IBoard](Protocol):
    """
    A class representing a Syzygy tablebase for chess endgame analysis.

    Attributes:
        table_base (chess.syzygy.Tablebase): The Syzygy tablebase object.

    Methods:
        fast_in_table(board: T_Board) -> bool:
            Check if the given board is suitable for fast tablebase lookup.

        in_table(board: T_Board) -> bool:
            Check if the given board is in the tablebase.

        get_over_event(board: T_Board) -> tuple[Winner, HowOver]:
            Get the winner and how the game is over for the given board.

        val(board: T_Board) -> int:
            Get the value of the given board from the tablebase.

        value_white(board: T_Board) -> int:
            Get the value of the given board for the white player.

        get_over_tag(board: T_Board) -> OverTags:
            Get the over tag for the given board.

        string_result(board: T_Board) -> str:
            Get the string representation of the result for the given board.

        dtz(board: T_Board) -> int:
            Get the distance-to-zero (DTZ) value for the given board.

        best_move(board: T_Board) -> moveKey:
            Get the best move according to the tablebase for the given board.
    """

    def fast_in_table(self, board: T_Board) -> bool:
        """
        Check if the given board is suitable for fast tablebase lookup.

        Args:
            board (boards.BoardChi): The board to check.

        Returns:
            bool: True if the board is suitable for fast lookup, False otherwise.
        """
        return board.number_of_pieces_on_the_board() < 6

    def in_table(self, board: T_Board) -> bool:
        """
        Check if the given board is in the tablebase.

        Args:
            board (boards.BoardChi): The board to check.

        Returns:
            bool: True if the board is in the tablebase, False otherwise.
        """
        try:
            self.wdl(board=board)
        except KeyError:
            return False
        return True

    def wdl(self, board: T_Board) -> int:
        """
        Probe the tablebase for the win/draw/loss value of the given board.

        Args:
            board (T_Board): The board to probe.

        Returns:
            int: 2 if the side to move is winning, 0 if draw, -2 if losing.
        """
        # Example implementation, replace with actual tablebase probing logic
        # For now, raise NotImplementedError to indicate it must be implemented
        raise NotImplementedError("wdl method must be implemented in subclass")

    def get_over_event(self, board: T_Board) -> tuple[Winner, HowOver]:
        """
        Get the winner and how the game is over for the given board.

        Args:
            board (boards.BoardChi): The board to analyze.

        Returns:
            tuple[Winner, HowOver]: The winner and how the game is over.
        """
        val: int = self.val(board)

        who_is_winner_: Winner = Winner.NO_KNOWN_WINNER
        how_over_: HowOver
        if val != 0:
            how_over_ = HowOver.WIN
            if val > 0:
                who_is_winner_ = (
                    Winner.WHITE if board.turn == chess.WHITE else Winner.BLACK
                )
            if val < 0:
                who_is_winner_ = (
                    Winner.WHITE if board.turn == chess.BLACK else Winner.BLACK
                )
        else:
            how_over_ = HowOver.DRAW

        return who_is_winner_, how_over_

    def val(self, board: T_Board) -> int:
        """
        Get the value of the given board from the tablebase.

        Args:
            board (boards.BoardChi): The board to get the value for.

        Returns:
            int: The value of the board from the tablebase.
        """
        # tablebase.probe_wdl Returns 2 if the side to move is winning, 0 if the position is a draw and -2 if the side to move is losing.
        val: int = self.wdl(board)
        return val

    def value_white(self, board: T_Board) -> int:
        """
        Get the value of the given board for the white player.

        Args:
            board (boards.BoardChi): The board to get the value for.

        Returns:
            int: The value of the board for the white player.
        """
        # tablebase.probe_wdl Returns 2 if the side to move is winning, 0 if the position is a draw and -2 if the side to move is losing.
        val: int = self.wdl(board)
        if board.turn == chess.WHITE:
            return val * 100000
        else:
            return val * -10000

    def get_over_tag(self, board: T_Board) -> OverTags:
        """
        Get the over tag for the given board.

        Args:
            board (boards.BoardChi): The board to get the over tag for.

        Returns:
            OverTags: The over tag for the board.
        """
        val = self.wdl(board)
        if val > 0:
            if board.turn == chess.WHITE:
                return OverTags.TAG_WIN_WHITE
            else:
                return OverTags.TAG_WIN_BLACK
        elif val == 0:
            return OverTags.TAG_DRAW
        else:
            if board.turn == chess.WHITE:
                return OverTags.TAG_WIN_BLACK
            else:
                return OverTags.TAG_WIN_WHITE

    def string_result(self, board: T_Board) -> str:
        """
        Get the string representation of the result for the given board.

        Args:
            board (boards.BoardChi): The board to get the result for.

        Returns:
            str: The string representation of the result.
        """
        val = self.wdl(board)
        player_to_move = "white" if board.turn == chess.WHITE else "black"
        if val > 0:
            return "WIN for player " + player_to_move
        elif val == 0:
            return "DRAW"
        else:
            return "LOSS for player " + player_to_move

    def dtz(self, board: T_Board) -> int:
        """
        Get the distance-to-zero (DTZ) value for the given board.

        Args:
            board (boards.BoardChi): The board to get the DTZ value for.

        Returns:
            int: The DTZ value for the board.
        """
        # Example implementation, replace with actual tablebase probing logic
        # For now, return a default value (e.g., 0)
        raise NotImplementedError("wdl method must be implemented in subclass")

    def best_move(self, board: T_Board) -> moveKey:
        """
        Get the best move according to the tablebase for the given board.

        Args:
            board (boards.BoardChi): The board to find the best move for.

        Returns:
            chess.Move: The best move according to the tablebase.
        """
        all_moves: list[moveKey] = board.legal_moves.get_all()
        # avoid draws by 50 move rules in winning position, # otherwise look
        # for it to make it last and preserve pieces in case of mistake by opponent

        best_value = -1000000000000000000000

        assert all_moves
        best_move: moveKey = all_moves[0]
        move: moveKey
        for move in all_moves:
            board_copy: T_Board = board.copy(stack=True)
            board_copy.play_move_key(move=move)
            val_player_next_board = self.val(board_copy)
            val_player_node = -val_player_next_board
            dtz_player_next_board = self.dtz(board_copy)
            dtz_player_node = -dtz_player_next_board
            if val_player_node > 0:  # winning position
                new_value = (
                    board.is_zeroing(move) * 100
                    - dtz_player_node * (1 - int(board.is_zeroing(move)))
                    + 10000
                )
            elif val_player_node == 0:  # drawing position
                new_value = -board.is_zeroing(move) * 100 + dtz_player_node * (
                    1 - int(board.is_zeroing(move))
                )
            else:  # losing position
                new_value = (
                    -board.is_zeroing(move) * 100
                    - dtz_player_node * (1 - int(board.is_zeroing(move)))
                    - 10000
                )

            if new_value > best_value:
                best_value = new_value
                best_move = move

        return best_move



#! /usr/bin/env python

"""
This module is the execution point of the chess GUI application.

It provides the `MainWindow` class, which creates a surface for the chessboard and handles user interactions.
"""

import math
import os
import queue
import time
import typing

import chess
import chess.svg
import PySide6.QtGui as QtGui
from PySide6.QtCore import Qt, QTimer, Slot
from PySide6.QtGui import QIcon
from PySide6.QtSvgWidgets import QSvgWidget
from PySide6.QtWidgets import (
    QDialog,
    QLabel,
    QProgressBar,
    QPushButton,
    QTableWidget,
    QTableWidgetItem,
    QWidget,
)

from chipiron.environments.chess_env.board import BoardFactory, IBoard, create_board_chi
from chipiron.environments.chess_env.board.utils import FenPlusHistory
from chipiron.environments.chess_env.move import moveUci
from chipiron.games.game.game_playing_status import PlayingStatus
from chipiron.games.match.match_results import MatchResults, SimpleResults
from chipiron.players import PlayerFactoryArgs
from chipiron.players.move_selector.treevalue import TreeAndValuePlayerArgs
from chipiron.players.move_selector.treevalue.progress_monitor.progress_monitor import (
    TreeMoveLimitArgs,
)
from chipiron.players.player_ids import PlayerConfigTag
from chipiron.utils.communication.gui_messages import (
    BackMessage,
    EvaluationMessage,
    GameStatusMessage,
    PlayerProgressMessage,
)
from chipiron.utils.communication.gui_messages.gui_messages import MatchResultsMessage
from chipiron.utils.communication.gui_player_message import PlayersColorToPlayerMessage
from chipiron.utils.communication.player_game_messages import BoardMessage, MoveMessage
from chipiron.utils.dataclass import IsDataclass
from chipiron.utils.logger import chipiron_logger
from chipiron.utils.path_variables import GUI_DIR

if typing.TYPE_CHECKING:
    from chipiron.environments.chess_env.move.imove import moveKey


class MainWindow(QWidget):
    """
    Create a surface for the chessboard and handle user interactions.

    This class provides the main window for the chess GUI application. It handles user interactions such as
    clicking on chess pieces, making moves, and controlling the game status.

    Attributes:
        playing_status (PlayingStatus): The current playing status of the game.
        gui_mailbox (queue.Queue[IsDataclass]): The mailbox for receiving messages from the GUI thread.
        main_thread_mailbox (queue.Queue[IsDataclass]): The mailbox for sending messages to the main thread.
    """

    def __init__(
        self,
        gui_mailbox: queue.Queue[IsDataclass],
        main_thread_mailbox: queue.Queue[IsDataclass],
        board_factory: BoardFactory,
    ) -> None:
        """
        Initialize the chessboard and the main window.

        Args:
            gui_mailbox (queue.Queue[IsDataclass]): The mailbox for receiving messages from the GUI thread.
            main_thread_mailbox (queue.Queue[IsDataclass]): The mailbox for sending messages to the main thread.
        """
        super().__init__()

        self.play_button_clicked_last_time: float | None = None
        self.pause_button_clicked_last_time: float | None = None

        self.board_factory = board_factory
        self.playing_status = PlayingStatus.PLAY

        self.gui_mailbox = gui_mailbox
        self.main_thread_mailbox = main_thread_mailbox

        # Set window icon with existence check
        window_icon_path = os.path.join(GUI_DIR, "chipicon.png")
        if os.path.exists(window_icon_path):
            self.setWindowIcon(QIcon(window_icon_path))
        else:
            chipiron_logger.warning("Window icon file not found: %s", window_icon_path)

        self.setWindowTitle("🐙 ♛  Chipiron Chess GUI  ♛ 🐙")
        self.setGeometry(300, 300, 1400, 800)

        self.widgetSvg = QSvgWidget(parent=self)
        self.widgetSvg.setGeometry(10, 10, 600, 600)

        self.closeButton = QPushButton(self)
        self.closeButton.setText("Close")  # text
        self._check_and_set_icon(
            self.closeButton, os.path.join(GUI_DIR, "close.png")
        )  # icon
        self.closeButton.setShortcut("Ctrl+D")  # shortcut key
        self.closeButton.clicked.connect(self.stopppy)
        self.closeButton.setToolTip("Close the widget")  # Tool tip
        self.closeButton.move(800, 20)

        # self.play_button = QPushButton(self)
        # self.play_button.setText("Play")  # text
        # self.play_button.setIcon(QIcon(os.path.join(GUI_DIR, "play.png")))  # icon
        # self.play_button.clicked.connect(self.play_button_clicked)
        # self.play_button.setToolTip("play the game")  # Tool tip
        # self.play_button.move(700, 100)

        self.pause_button = QPushButton(self)
        self.pause_button.setText("Pause")  # text
        self._check_and_set_icon(
            self.pause_button, os.path.join(GUI_DIR, "pause.png")
        )  # icon
        self.pause_button.clicked.connect(self.pause_button_clicked)
        self.pause_button.setToolTip("pause the game")  # Tool tip
        self.pause_button.move(700, 100)

        self.back_button = QPushButton(self)
        self.back_button.setText("Back")  # text
        self._check_and_set_icon(
            self.back_button, os.path.join(GUI_DIR, "back.png")
        )  # icon
        self.back_button.clicked.connect(self.back_button_clicked)
        self.back_button.setToolTip("back one move")  # Tool tip
        self.back_button.move(900, 100)

        self.player_white_button = QPushButton(self)
        self.player_white_button.setText("Player")  # text
        self._check_and_set_icon(
            self.player_white_button, os.path.join(GUI_DIR, "white_king.png")
        )  # icon
        self.player_white_button.setStyleSheet(
            "QPushButton {background-color: white; color: black;}"
        )
        self.player_white_button.setGeometry(620, 200, 470, 30)

        self.progress_white = QProgressBar(self)
        self.progress_white.setGeometry(620, 230, 470, 30)

        self.player_black_button = QPushButton(self)
        self.player_black_button.setText("Player")  # text
        self._check_and_set_icon(
            self.player_black_button, os.path.join(GUI_DIR, "black_king.png")
        )  # icon
        self.player_black_button.setStyleSheet(
            "QPushButton {background-color: black; color: white;}"
        )
        self.player_black_button.setGeometry(620, 300, 470, 30)

        self.progress_black = QProgressBar(self)
        self.progress_black.setGeometry(620, 330, 470, 30)

        self.tablewidget = QTableWidget(1, 2, self)
        self.tablewidget.setGeometry(1100, 200, 260, 330)

        self.score_button = QPushButton(self)
        self.score_button.setText("⚖ Score 0-0")  # text
        self.score_button.setStyleSheet(
            "QPushButton {background-color: white; color: black;}"
        )
        self.score_button.setGeometry(620, 400, 370, 30)

        self.round_button = QPushButton(self)
        self.round_button.setText("🎲 Round")  # text
        self.round_button.setStyleSheet(
            "QPushButton {background-color: white; color: black;}"
        )
        self.round_button.setGeometry(620, 500, 370, 30)

        self.fen_button = QLabel(self)
        self.fen_button.setText("fen")  # text
        self.fen_button.setStyleSheet(
            "QPushButton { background-color: white; color: black;}"
        )
        self.fen_button.setGeometry(20, 620, 550, 30)
        self.fen_button.setTextInteractionFlags(
            QtGui.Qt.TextInteractionFlag.TextSelectableByMouse
        )

        self.legal_moves_button = QLabel(self)
        self.legal_moves_button.setText("legal moves")  # text
        self.legal_moves_button.setStyleSheet(
            "QPushButton { background-color: white; color: black;}"
        )
        self.legal_moves_button.setGeometry(20, 650, 550, 80)
        self.legal_moves_button.setTextInteractionFlags(
            QtGui.Qt.TextInteractionFlag.TextSelectableByMouse
        )

        self.eval_button = QPushButton(self)
        self.eval_button.setText("🐟 Eval")  # text
        self.eval_button.setStyleSheet(
            "QPushButton {background-color: white; color: black;}"
        )
        self.eval_button.setGeometry(620, 600, 470, 30)

        self.eval_button_chi = QPushButton(self)
        self.eval_button_chi.setText("🐙 Eval")  # text
        self.eval_button_chi.setStyleSheet(
            "QPushButton {background-color: white; color: black;}"
        )
        self.eval_button_chi.setGeometry(620, 650, 470, 30)

        self.eval_button_white = QPushButton(self)
        self.eval_button_white.setText("♕ White Eval")  # text
        self.eval_button_white.setStyleSheet(
            "QPushButton {background-color: white; color: black;}"
        )
        self.eval_button_white.setGeometry(620, 700, 470, 30)

        self.eval_button_black = QPushButton(self)
        self.eval_button_black.setText("♛ Black Eval")  # text
        self.eval_button_black.setStyleSheet(
            "QPushButton {background-color: white; color: black;}"
        )
        self.eval_button_black.setGeometry(620, 750, 470, 30)

        self.board_size = min(self.widgetSvg.width(), self.widgetSvg.height())
        self.coordinates = True
        self.margin = 0.05 * self.board_size if self.coordinates else 0
        self.squareSize = (self.board_size - 2 * self.margin) / 8.0
        self.pieceToMove = [None, None]

        self.checkThreadTimer = QTimer(self)
        self.checkThreadTimer.setInterval(5)  # .5 seconds
        self.checkThreadTimer.timeout.connect(self.process_message)
        self.checkThreadTimer.start()

    def _check_and_set_icon(self, button: QPushButton, icon_path: str) -> None:
        """
        Check if icon file exists and set it, otherwise log a warning.

        Args:
            button: The button to set the icon for
            icon_path: The path to the icon file
        """
        if os.path.exists(icon_path):
            button.setIcon(QIcon(icon_path))
        else:
            chipiron_logger.warning("Icon file not found: %s", icon_path)
            # Set a default icon or leave empty
            button.setIcon(QIcon())  # Empty icon

    def stopppy(self) -> None:
        """
        Stops the execution of the GUI.

        This method closes the GUI window and sends a kill message to the main thread.

        Returns:
            None
        """
        # should we send a kill message to the main thread?
        self.close()

    def play_button_clicked(self) -> None:
        """
        Handle the event when the play button is clicked.

        This method prints a message indicating that the play button has been clicked,
        and sends a GameStatusMessage with the status set to PlayingStatus.PLAY to the main thread mailbox.
        """
        if self.playing_status == PlayingStatus.PAUSE:
            if (
                self.play_button_clicked_last_time is None
                or abs(self.play_button_clicked_last_time - time.time()) > 0.01
            ):
                chipiron_logger.info("play_button_clicked")
                message: GameStatusMessage = GameStatusMessage(
                    status=PlayingStatus.PLAY
                )
                self.main_thread_mailbox.put(message)
                self.play_button_clicked_last_time = time.time()

    def back_button_clicked(self) -> None:
        """
        Handle the event when the back button is clicked.

        This method prints a message and puts a `BackMessage` object into the main thread mailbox.
        """
        chipiron_logger.info("back_button_clicked")
        message: BackMessage = BackMessage()
        self.main_thread_mailbox.put(message)

    def pause_button_clicked(self) -> None:
        """
        Handles the click event of the pause button.

        Prints 'pause_button_clicked' and sends a GameStatusMessage with the status set to PlayingStatus.PAUSE
        to the main thread mailbox.
        """
        if self.playing_status == PlayingStatus.PLAY:
            if (
                self.pause_button_clicked_last_time is None
                or abs(self.pause_button_clicked_last_time - time.time()) > 0.01
            ):
                chipiron_logger.info("pause_button_clicked")
                message: GameStatusMessage = GameStatusMessage(
                    status=PlayingStatus.PAUSE
                )
                self.main_thread_mailbox.put(message)
                self.pause_button_clicked_last_time = time.time()

    @typing.no_type_check
    @Slot(QWidget)
    def mousePressEvent(self, event):
        """
        Handle left mouse clicks and enable moving chess pieces by
        clicking on a chess piece and then the target square.

        Moves must be made according to the rules of chess because
        illegal moves are suppressed.
        """
        if event.x() <= self.board_size and event.y() <= self.board_size:
            if event.buttons() == Qt.LeftButton:
                if (
                    self.margin < event.x() < self.board_size - self.margin
                    and self.margin < event.y() < self.board_size - self.margin
                ):
                    file = int((event.x() - self.margin) / self.squareSize)
                    rank = 7 - int((event.y() - self.margin) / self.squareSize)
                    square = chess.square(file, rank)
                    piece = self.board.piece_at(square)
                    self.coordinates = f"{chr(file + 97)}{rank + 1}"
                    if self.pieceToMove[0] is not None:
                        try:
                            all_moves_keys: list[moveKey] = (
                                self.board.legal_moves.get_all()
                            )
                            all_legal_moves_uci: list[moveUci] = [
                                self.board.legal_moves.generated_moves[move_key].uci()
                                for move_key in all_moves_keys
                            ]
                            move: chess.Move = chess.Move.from_uci(
                                "{}{}".format(self.pieceToMove[1], self.coordinates)
                            )
                            move_promote: chess.Move = chess.Move.from_uci(
                                "{}{}q".format(self.pieceToMove[1], self.coordinates)
                            )
                            if move.uci() in all_legal_moves_uci:
                                self.send_move_to_main_thread(move_uci=move.uci())
                            elif move_promote.uci() in all_legal_moves_uci:
                                self.choice_promote()
                                self.send_move_to_main_thread(
                                    move_uci=self.move_promote_asked.uci()
                                )
                            else:
                                legal_moves_uci: list[moveUci] = [
                                    self.board.get_uci_from_move_key(move_key)
                                    for move_key in all_moves_keys
                                ]
                                chipiron_logger.info(
                                    "Looks like the move %s is a wrong move.. "
                                    "The legals moves are %s in %s",
                                    move,
                                    legal_moves_uci,
                                    self.board,
                                )
                        except ValueError:
                            chipiron_logger.info("Oops!  Doubleclicked?  Try again...")
                        piece = None
                    self.pieceToMove = [piece, self.coordinates]

    def send_move_to_main_thread(self, move_uci: moveUci) -> None:
        """
        Sends a move to the main thread for processing.

        Args:
            move (chess.Move): The move to be sent.

        Returns:
            None
        """
        message: MoveMessage = MoveMessage(
            move=move_uci,
            corresponding_board=self.board.fen,
            player_name=PlayerConfigTag.GUI_HUMAN,
            color_to_play=self.board.turn,
        )
        self.main_thread_mailbox.put(item=message)

    @typing.no_type_check
    def choice_promote(self):
        """
        Displays a dialog box with buttons for promoting a chess piece.

        The dialog box allows the user to choose between promoting the pawn to a queen, rook, bishop, or knight.
        Each button is connected to a corresponding method for handling the promotion.

        Returns:
            None
        """

        self.d = QDialog()
        d = self.d
        d.setWindowTitle("Promote to ?")
        d.setWindowModality(Qt.ApplicationModal)

        d.closeButtonQ = QPushButton(d)
        d.closeButtonQ.setText("Queen")  # text
        d.closeButtonQ.setStyleSheet(
            "QPushButton {background-color: white; color: blue;}"
        )
        d.closeButtonQ.setGeometry(150, 100, 150, 20)
        d.closeButtonQ.clicked.connect(self.promote_queen)

        d.closeButtonR = QPushButton(d)
        d.closeButtonR.setText("Rook")  # text
        d.closeButtonR.setStyleSheet(
            "QPushButton {background-color: white; color: blue;}"
        )
        d.closeButtonR.setGeometry(150, 200, 150, 20)
        d.closeButtonR.clicked.connect(self.promote_rook)

        d.closeButtonB = QPushButton(d)
        d.closeButtonB.setText("Bishop")  # text
        d.closeButtonB.setStyleSheet(
            "QPushButton {background-color: white; color: blue;}"
        )
        d.closeButtonB.setGeometry(150, 300, 150, 20)
        d.closeButtonB.clicked.connect(self.promote_bishop)

        d.closeButtonK = QPushButton(d)
        d.closeButtonK.setText("Knight")  # text
        d.closeButtonK.setStyleSheet(
            "QPushButton {background-color: white; color: blue;}"
        )
        d.closeButtonK.setGeometry(150, 400, 150, 20)
        d.closeButtonK.clicked.connect(self.promote_knight)

        d.exec_()

    @typing.no_type_check
    def promote_queen(self):
        """
        Promotes the selected piece to a queen.

        This method creates a move object to promote the selected piece to a queen by appending 'q' to the UCI notation
        of the piece's destination square. It then closes the dialog window.

        Returns:
            None
        """
        self.move_promote_asked = chess.Move.from_uci(
            "{}{}q".format(self.pieceToMove[1], self.coordinates)
        )
        self.d.close()

    @typing.no_type_check
    def promote_rook(self):
        """
        Promotes a pawn to a rook.

        This method is called when a pawn reaches the opposite end of the board and needs to be promoted to a rook.
        It creates a move object representing the promotion and closes the dialog window.

        Returns:
            None
        """
        self.move_promote_asked = chess.Move.from_uci(
            f"{self.pieceToMove[1]}{self.coordinates}r"
        )
        self.d.close()

    @typing.no_type_check
    def promote_bishop(self):
        """
        Promotes the current piece to a bishop.

        This method creates a move object to promote the current piece to a bishop and closes the dialog window.

        Returns:
            None
        """
        self.move_promote_asked = chess.Move.from_uci(
            f"{self.pieceToMove[1]}{self.coordinates}b"
        )
        self.d.close()

    @typing.no_type_check
    def promote_knight(self):
        """
        Promotes a pawn to a knight.

        This method is called when a pawn is promoted to a knight in the GUI.
        It creates a move object representing the promotion and closes the GUI.

        Returns:
        None
        """
        self.move_promote_asked = chess.Move.from_uci(
            f"{self.pieceToMove[1]}{self.coordinates}n"
        )
        self.d.close()

    def process_message(self) -> None:
        """
        Process a message received by the GUI.

        Draw a chessboard with the starting position and then redraw
        it for every new move.

        This method is responsible for handling different types of messages
        received by the GUI and taking appropriate actions based on the message type.

        Supported message types:
        - BoardMessage: Updates the board and redraws it.
        - EvaluationMessage: Updates the evaluation values.
        - PlayersColorToPlayerMessage: Updates the mapping of player colors to player information.
        - MatchResultsMessage: Updates the match results.
        - GameStatusMessage: Updates the game play status.
        - Other: Raises a ValueError indicating an unknown message type.

        Returns:
        None
        """
        if not self.gui_mailbox.empty():
            message = self.gui_mailbox.get()
            match message:
                case BoardMessage():
                    board_message: BoardMessage = message
                    self.board: IBoard = self.board_factory(
                        fen_with_history=board_message.fen_plus_moves
                    )
                    chipiron_logger.info("GUI receiving board %s", self.board.fen)
                    self.draw_board()
                    self.display_move_history()
                case PlayerProgressMessage():
                    progress_message: PlayerProgressMessage = message
                    if (
                        progress_message.player_color == chess.WHITE
                        and progress_message.progress_percent is not None
                    ):
                        self.progress_white.setValue(progress_message.progress_percent)
                    if (
                        progress_message.player_color == chess.BLACK
                        and progress_message.progress_percent is not None
                    ):
                        self.progress_black.setValue(progress_message.progress_percent)
                case EvaluationMessage():
                    evaluation_message: EvaluationMessage = message
                    evaluation_stock = evaluation_message.evaluation_stock
                    evaluation_chipiron = evaluation_message.evaluation_chipiron
                    evaluation_black = evaluation_message.evaluation_player_black
                    evaluation_white = evaluation_message.evaluation_player_white
                    self.update_evaluation(
                        evaluation_stock=evaluation_stock,
                        evaluation_chipiron=evaluation_chipiron,
                        evaluation_white=evaluation_white,
                        evaluation_black=evaluation_black,
                    )
                case PlayersColorToPlayerMessage():
                    player_color_message: PlayersColorToPlayerMessage = message
                    players_color_to_player: dict[chess.Color, PlayerFactoryArgs] = (
                        player_color_message.player_color_to_factory_args
                    )
                    self.update_players_color_to_id(players_color_to_player)
                case MatchResultsMessage():
                    match_message: MatchResultsMessage = message
                    match_results: MatchResults = match_message.match_results
                    self.update_match_stats(match_results)
                case GameStatusMessage():
                    chipiron_logger.info("GameStatusMessage %s", message)
                    game_status_message: GameStatusMessage = message
                    play_status: PlayingStatus = game_status_message.status
                    self.update_game_play_status(play_status)
                case other:
                    raise ValueError(
                        f"unknown type of message received by gui {other} in {__name__}"
                    )

    def display_move_history(self) -> None:
        """
        Display the move history in a table widget.

        This method calculates the number of rounds based on the number of half moves in the move stack.
        It then sets the number of rows in the table widget to the number of rounds.
        The table widget's horizontal header labels are set to 'White' and 'Black'.
        The move history is then iterated over and each move is added to the table widget.

        Returns:
            None
        """

        num_half_move: int = len(self.board.move_history_stack)
        num_rounds: int = int(math.ceil(num_half_move / 2))
        self.tablewidget.setRowCount(num_rounds)
        self.tablewidget.setHorizontalHeaderLabels(["White", "Black"])
        for player in range(2):
            for round_ in range(num_rounds):
                half_move = round_ * 2 + player
                if half_move < num_half_move:
                    item = QTableWidgetItem(
                        str(self.board.move_history_stack[half_move])
                    )
                    self.tablewidget.setItem(round_, player, item)

    def draw_board(self) -> None:
        """
        Draw a chessboard with the starting position and then redraw
        it for every new move.

        Returns:
            None
        """
        board_chi = create_board_chi(
            fen_with_history=FenPlusHistory(
                current_fen=self.board.fen,
                historical_moves=self.board.move_history_stack,
            )
        )

        repr_svg: str = chess.svg.board(
            board=board_chi.chess_board,
            size=390,
            lastmove=board_chi.chess_board.peek()
            if board_chi.chess_board.move_stack
            else None,
            check=board_chi.chess_board.king(board_chi.chess_board.turn)
            if board_chi.chess_board.is_check()
            else None,
        )

        self.boardSvg = repr_svg.encode("UTF-8")
        self.drawBoardSvg = self.widgetSvg.load(self.boardSvg)
        self.round_button.setText(
            "🎲 Round: " + str(self.board.fullmove_number)
        )  # text
        self.fen_button.setText("🔧 <b>fen:</b> " + str(self.board.fen))  # text

        all_moves_keys_chi = self.board.legal_moves.get_all()
        all_moves_uci_chi = [
            self.board.get_uci_from_move_key(move_key=move_key)
            for move_key in all_moves_keys_chi
        ]

        self.legal_moves_button.setWordWrap(True)
        self.legal_moves_button.setMinimumHeight(100)
        lines: list[str] = []

        # Only add sublists if they are non-empty
        for sublist in [
            all_moves_uci_chi[:7],
            all_moves_uci_chi[7:14],
            all_moves_uci_chi[14:21],
            all_moves_uci_chi[21:28],
        ]:
            if sublist:  # skip empty
                lines.append("    " + str(sublist))

        # Join lines with <br>
        moves_html = "\n".join(lines)

        self.legal_moves_button.setText(
            f"📋 <b>legal moves:</b><pre>{moves_html}</pre>"
        )
        return self.drawBoardSvg

    def extract_message_from_player(self, player: PlayerFactoryArgs) -> str:
        """
        Extracts a message from a player to be shown in the GUI.

        Args:
            player (PlayerFactoryArgs): The factory arguments for the player.

        Returns:
            str: The extracted message.
        """
        name: str = player.player_args.name
        tree_move_limit: str | int = ""
        if isinstance(player.player_args.main_move_selector, TreeAndValuePlayerArgs):
            if isinstance(
                player.player_args.main_move_selector.stopping_criterion,
                TreeMoveLimitArgs,
            ):
                tree_move_limit = player.player_args.main_move_selector.stopping_criterion.tree_move_limit

        return f"{name} ({tree_move_limit})"

    def update_players_color_to_id(
        self, players_color_to_player: dict[chess.Color, PlayerFactoryArgs]
    ) -> None:
        """
        Update the player buttons with the corresponding player names.

        Args:
            players_color_to_player (dict[chess.Color, str]): A dictionary mapping chess.Color to player names.

        Returns:
            None
        """
        self.player_white_button.setText(
            " White: "
            + self.extract_message_from_player(players_color_to_player[chess.WHITE])
        )  # text
        self.player_black_button.setText(
            " Black: "
            + self.extract_message_from_player(players_color_to_player[chess.BLACK])
        )  # text

        if players_color_to_player[chess.BLACK].player_args.is_human():
            self.progress_black.setTextVisible(True)
            self.progress_black.setValue(100)
            self.progress_black.setFormat("Think Human!")
        else:
            self.progress_black.setValue(0)

        if players_color_to_player[chess.WHITE].player_args.is_human():
            self.progress_white.setTextVisible(True)
            self.progress_white.setValue(100)
            self.progress_white.setFormat("Think Human!")
        else:
            self.progress_white.setValue(0)

    def update_evaluation(
        self,
        evaluation_stock: float,
        evaluation_chipiron: float,
        evaluation_white: float,
        evaluation_black: float,
    ) -> None:
        """
        Update the evaluation values displayed on the GUI.

        Args:
            evaluation_stock (float): The evaluation value for the stock.
            evaluation_chipiron (float): The evaluation value for the chipiron.
            evaluation_white (float): The evaluation value for the white.
            evaluation_black (float): The evaluation value for the black.

        Returns:
            None
        """
        self.eval_button.setText("📊 Eval 🐟: " + str(evaluation_stock))  # text
        self.eval_button_chi.setText("🧮 Eval 🐙: " + str(evaluation_chipiron))  # text
        self.eval_button_black.setText(
            "🧠 Eval White: " + str(evaluation_white)
        )  # text
        self.eval_button_white.setText(
            "🧠 Eval Black: " + str(evaluation_black)
        )  # text

    def update_game_play_status(self, play_status: PlayingStatus) -> None:
        """Update the game play status.

        Args:
            play_status (PlayingStatus): The new playing status.
        """
        chipiron_logger.info("update_game_play_status %s", play_status)

        if self.playing_status != play_status:
            self.playing_status = play_status
            if play_status == PlayingStatus.PAUSE:
                self.pause_button.setText("Play")  # text
                self._check_and_set_icon(
                    self.pause_button, os.path.join(GUI_DIR, "play.png")
                )  # icon
                self.pause_button.clicked.connect(self.play_button_clicked)
                self.pause_button.setToolTip("play the game")  # Tool tip
                self.pause_button.move(700, 100)
            elif play_status == PlayingStatus.PLAY:
                self.pause_button.setText("Pause")  # text
                self._check_and_set_icon(
                    self.pause_button, os.path.join(GUI_DIR, "pause.png")
                )  # icon
                self.pause_button.clicked.connect(self.pause_button_clicked)
                self.pause_button.setToolTip("pause the game")  # Tool tip
                self.pause_button.move(700, 100)

    def update_match_stats(self, match_result: MatchResults) -> None:
        """
        Update the match statistics and display them on the GUI.

        Args:
            match_result (MatchResults): The result of the match.

        Returns:
            None
        """
        simple_results: SimpleResults = match_result.get_simple_result()
        self.score_button.setText(
            "⚖ Score: "
            + str(simple_results.player_one_wins)
            + "-"
            + str(simple_results.player_two_wins)
            + "-"
            + str(simple_results.draws)
        )  # text

        chipiron_logger.info("update %s", match_result.match_finished)
        # if the match is over we kill the GUI
        if match_result.match_finished:
            chipiron_logger.info("finishing the widget")
            self.close()
